# -*- coding: utf-8 -*-

from collections import namedtuple
from functools import partial
from operator import index

import jax
import numpy as np
from jax import lax, jit, vmap, numpy as jnp, random as jr, core
from jax._src import dtypes
from jax.experimental.host_callback import call
from jax.tree_util import register_pytree_node

from brainpy.math.jaxarray import JaxArray, Variable
from brainpy.tools.errors import check_error_in_jit
from .utils import wraps

__all__ = [
  'RandomState', 'Generator',

  'seed', 'default_rng',

  'rand', 'randint', 'random_integers', 'randn', 'random',
  'random_sample', 'ranf', 'sample', 'choice', 'permutation', 'shuffle', 'beta',
  'exponential', 'gamma', 'gumbel', 'laplace', 'logistic', 'normal', 'pareto',
  'poisson', 'standard_cauchy', 'standard_exponential', 'standard_gamma',
  'standard_normal', 'standard_t', 'uniform', 'truncated_normal', 'bernoulli',
  'lognormal', 'binomial', 'chisquare', 'dirichlet', 'geometric', 'f',
  'hypergeometric', 'logseries', 'multinomial', 'multivariate_normal',
  'negative_binomial', 'noncentral_chisquare', 'noncentral_f', 'power',
  'rayleigh', 'triangular', 'vonmises', 'wald', 'weibull', 'weibull_min',
  'zipf', 'maxwell', 't', 'orthogonal', 'loggamma', 'categorical',
]


def _size2shape(size):
  if size is None:
    return ()
  elif isinstance(size, int):
    return (size,)
  elif isinstance(size, (tuple, list)):
    return tuple(size)
  else:
    raise ValueError(f'Must be a list/tuple of int, but got {size}')


def _check_shape(name, shape, *param_shapes):
  shape = core.as_named_shape(shape)
  if param_shapes:
    shape_ = lax.broadcast_shapes(shape.positional, *param_shapes)
    if shape.positional != shape_:
      msg = ("{} parameter shapes must be broadcast-compatible with shape "
             "argument, and the result of broadcasting the shapes must equal "
             "the shape argument, but got result {} for shape argument {}.")
      raise ValueError(msg.format(name, shape_, shape))


def _remove_jax_array(a):
  return a.value if isinstance(a, JaxArray) else a


def _const(example, val):
  dtype = dtypes.dtype(example, canonicalize=True)
  if dtypes.is_python_scalar(example):
    val = dtypes.scalar_type_of(example)(val)
    return val if dtype == dtypes.dtype(val, canonicalize=True) else np.array(val, dtype)
  return np.array(val, dtype)


_tr_params = namedtuple(
  "tr_params", ["c", "b", "a", "alpha", "u_r", "v_r", "m", "log_p", "log1_p", "log_h"]
)


def _get_tr_params(n, p):
  # See Table 1. Additionally, we pre-compute log(p), log1(-p) and the
  # constant terms, that depend only on (n, p, m) in log(f(k)) (bottom of page 5).
  mu = n * p
  spq = jnp.sqrt(mu * (1 - p))
  c = mu + 0.5
  b = 1.15 + 2.53 * spq
  a = -0.0873 + 0.0248 * b + 0.01 * p
  alpha = (2.83 + 5.1 / b) * spq
  u_r = 0.43
  v_r = 0.92 - 4.2 / b
  m = jnp.floor((n + 1) * p).astype(n.dtype)
  log_p = jnp.log(p)
  log1_p = jnp.log1p(-p)
  log_h = (m + 0.5) * (jnp.log((m + 1.0) / (n - m + 1.0)) + log1_p - log_p) + (
      _stirling_approx_tail(m) + _stirling_approx_tail(n - m)
  )
  return _tr_params(c, b, a, alpha, u_r, v_r, m, log_p, log1_p, log_h)


def _stirling_approx_tail(k):
  precomputed = jnp.array([0.08106146679532726,
                           0.04134069595540929,
                           0.02767792568499834,
                           0.02079067210376509,
                           0.01664469118982119,
                           0.01387612882307075,
                           0.01189670994589177,
                           0.01041126526197209,
                           0.009255462182712733,
                           0.008330563433362871, ])
  kp1 = k + 1
  kp1sq = (k + 1) ** 2
  return jnp.where(
    k < 10, precomputed[k],
    (1.0 / 12 - (1.0 / 360 - (1.0 / 1260) / kp1sq) / kp1sq) / kp1,
  )


def _binomial_btrs(key, p, n):
  """
  Based on the transformed rejection sampling algorithm (BTRS) from the
  following reference:

  Hormann, "The Generation of Binonmial Random Variates"
  (https://core.ac.uk/download/pdf/11007254.pdf)
  """

  def _btrs_body_fn(val):
    _, key, _, _ = val
    key, key_u, key_v = jr.split(key, 3)
    u = jr.uniform(key_u)
    v = jr.uniform(key_v)
    u = u - 0.5
    k = jnp.floor(
      (2 * tr_params.a / (0.5 - jnp.abs(u)) + tr_params.b) * u + tr_params.c
    ).astype(n.dtype)
    return k, key, u, v

  def _btrs_cond_fn(val):
    def accept_fn(k, u, v):
      # See acceptance condition in Step 3. (Page 3) of TRS algorithm
      # v <= f(k) * g_grad(u) / alpha

      m = tr_params.m
      log_p = tr_params.log_p
      log1_p = tr_params.log1_p
      # See: formula for log(f(k)) at bottom of Page 5.
      log_f = (
          (n + 1.0) * jnp.log((n - m + 1.0) / (n - k + 1.0))
          + (k + 0.5) * (jnp.log((n - k + 1.0) / (k + 1.0)) + log_p - log1_p)
          + (_stirling_approx_tail(k) - _stirling_approx_tail(n - k))
          + tr_params.log_h
      )
      g = (tr_params.a / (0.5 - jnp.abs(u)) ** 2) + tr_params.b
      return jnp.log((v * tr_params.alpha) / g) <= log_f

    k, key, u, v = val
    early_accept = (jnp.abs(u) <= tr_params.u_r) & (v <= tr_params.v_r)
    early_reject = (k < 0) | (k > n)
    return lax.cond(
      early_accept | early_reject,
      (),
      lambda _: ~early_accept,
      (k, u, v),
      lambda x: ~accept_fn(*x),
    )

  tr_params = _get_tr_params(n, p)
  ret = lax.while_loop(
    _btrs_cond_fn, _btrs_body_fn, (-1, key, 1.0, 1.0)
  )  # use k=-1 initially so that cond_fn returns True
  return ret[0]


def _binomial_inversion(key, p, n):
  def _binom_inv_body_fn(val):
    i, key, geom_acc = val
    key, key_u = jr.split(key)
    u = jr.uniform(key_u)
    geom = jnp.floor(jnp.log1p(-u) / log1_p) + 1
    geom_acc = geom_acc + geom
    return i + 1, key, geom_acc

  def _binom_inv_cond_fn(val):
    i, _, geom_acc = val
    return geom_acc <= n

  log1_p = jnp.log1p(-p)
  ret = lax.while_loop(_binom_inv_cond_fn, _binom_inv_body_fn, (-1, key, 0.0))
  return ret[0]


def _binomial_dispatch(key, p, n):
  def dispatch(key, p, n):
    is_le_mid = p <= 0.5
    pq = jnp.where(is_le_mid, p, 1 - p)
    mu = n * pq
    k = lax.cond(
      mu < 10,
      (key, pq, n),
      lambda x: _binomial_inversion(*x),
      (key, pq, n),
      lambda x: _binomial_btrs(*x),
    )
    return jnp.where(is_le_mid, k, n - k)

  # Return 0 for nan `p` or negative `n`, since nan values are not allowed for integer types
  cond0 = jnp.isfinite(p) & (n > 0) & (p > 0)
  return lax.cond(
    cond0 & (p < 1),
    (key, p, n),
    lambda x: dispatch(*x),
    (),
    lambda _: jnp.where(cond0, n, 0),
  )


@partial(jit, static_argnums=(3,))
def _binomial(key, p, n, shape):
  shape = shape or lax.broadcast_shapes(jnp.shape(p), jnp.shape(n))
  # reshape to map over axis 0
  p = jnp.reshape(jnp.broadcast_to(p, shape), -1)
  n = jnp.reshape(jnp.broadcast_to(n, shape), -1)
  key = jr.split(key, jnp.size(p))
  if jax.default_backend() == "cpu":
    ret = lax.map(lambda x: _binomial_dispatch(*x), (key, p, n))
  else:
    ret = vmap(lambda *x: _binomial_dispatch(*x))(key, p, n)
  return jnp.reshape(ret, shape)


@partial(jit, static_argnums=(2,))
def _categorical(key, p, shape):
  # this implementation is fast when event shape is small, and slow otherwise
  # Ref: https://stackoverflow.com/a/34190035
  shape = shape or p.shape[:-1]
  s = jnp.cumsum(p, axis=-1)
  r = jr.uniform(key, shape=shape + (1,))
  return jnp.sum(s < r, axis=-1)


def _scatter_add_one(operand, indices, updates):
  return lax.scatter_add(
    operand,
    indices,
    updates,
    lax.ScatterDimensionNumbers(
      update_window_dims=(),
      inserted_window_dims=(0,),
      scatter_dims_to_operand_dims=(0,),
    ),
  )


def _reshape(x, shape):
  if isinstance(x, (int, float, np.ndarray, np.generic)):
    return np.reshape(x, shape)
  else:
    return jnp.reshape(x, shape)


def _promote_shapes(*args, shape=()):
  # adapted from lax.lax_numpy
  if len(args) < 2 and not shape:
    return args
  else:
    shapes = [jnp.shape(arg) for arg in args]
    num_dims = len(lax.broadcast_shapes(shape, *shapes))
    return [
      _reshape(arg, (1,) * (num_dims - len(s)) + s) if len(s) < num_dims else arg
      for arg, s in zip(args, shapes)
    ]


@partial(jit, static_argnums=(3, 4))
def _multinomial(key, p, n, n_max, shape=()):
  if jnp.shape(n) != jnp.shape(p)[:-1]:
    broadcast_shape = lax.broadcast_shapes(jnp.shape(n), jnp.shape(p)[:-1])
    n = jnp.broadcast_to(n, broadcast_shape)
    p = jnp.broadcast_to(p, broadcast_shape + jnp.shape(p)[-1:])
  shape = shape or p.shape[:-1]
  if n_max == 0:
    return jnp.zeros(shape + p.shape[-1:], dtype=jnp.result_type(int))
  # get indices from categorical distribution then gather the result
  indices = _categorical(key, p, (n_max,) + shape)
  # mask out values when counts is heterogeneous
  if jnp.ndim(n) > 0:
    mask = _promote_shapes(jnp.arange(n_max) < jnp.expand_dims(n, -1), shape=shape + (n_max,))[0]
    mask = jnp.moveaxis(mask, -1, 0).astype(indices.dtype)
    excess = jnp.concatenate([jnp.expand_dims(n_max - n, -1),
                              jnp.zeros(jnp.shape(n) + (p.shape[-1] - 1,))],
                             -1)
  else:
    mask = 1
    excess = 0
  # NB: we transpose to move batch shape to the front
  indices_2D = (jnp.reshape(indices * mask, (n_max, -1))).T
  samples_2D = vmap(_scatter_add_one, (0, 0, 0))(jnp.zeros((indices_2D.shape[0], p.shape[-1]), dtype=indices.dtype),
                                                 jnp.expand_dims(indices_2D, axis=-1),
                                                 jnp.ones(indices_2D.shape, dtype=indices.dtype))
  return jnp.reshape(samples_2D, shape + p.shape[-1:]) - excess


@partial(jit, static_argnums=(2, 3))
def _von_mises_centered(key, concentration, shape, dtype=jnp.float64):
  """Compute centered von Mises samples using rejection sampling from [1]_ with wrapped Cauchy proposal.

  Returns
  -------
  out: array_like
     centered samples from von Mises

  References
  ----------
  .. [1] Luc Devroye "Non-Uniform Random Variate Generation", Springer-Verlag, 1986;
         Chapter 9, p. 473-476. http://www.nrbook.com/devroye/Devroye_files/chapter_nine.pdf

  """
  shape = shape or jnp.shape(concentration)
  dtype = jnp.result_type(dtype)
  concentration = lax.convert_element_type(concentration, dtype)
  concentration = jnp.broadcast_to(concentration, shape)

  s_cutoff_map = {
    jnp.dtype(jnp.float16): 1.8e-1,
    jnp.dtype(jnp.float32): 2e-2,
    jnp.dtype(jnp.float64): 1.2e-4,
  }
  s_cutoff = s_cutoff_map.get(dtype)

  r = 1.0 + jnp.sqrt(1.0 + 4.0 * concentration ** 2)
  rho = (r - jnp.sqrt(2.0 * r)) / (2.0 * concentration)
  s_exact = (1.0 + rho ** 2) / (2.0 * rho)

  s_approximate = 1.0 / concentration

  s = jnp.where(concentration > s_cutoff, s_exact, s_approximate)

  def cond_fn(*args):
    """check if all are done or reached max number of iterations"""
    i, _, done, _, _ = args[0]
    return jnp.bitwise_and(i < 100, jnp.logical_not(jnp.all(done)))

  def body_fn(*args):
    i, key, done, _, w = args[0]
    uni_ukey, uni_vkey, key = jr.split(key, 3)
    u = jr.uniform(
      key=uni_ukey,
      shape=shape,
      dtype=concentration.dtype,
      minval=-1.0,
      maxval=1.0,
    )
    z = jnp.cos(jnp.pi * u)
    w = jnp.where(done, w, (1.0 + s * z) / (s + z))  # Update where not done
    y = concentration * (s - w)
    v = jr.uniform(key=uni_vkey, shape=shape, dtype=concentration.dtype)
    accept = (y * (2.0 - y) >= v) | (jnp.log(y / v) + 1.0 >= y)
    return i + 1, key, accept | done, u, w

  init_done = jnp.zeros(shape, dtype=bool)
  init_u = jnp.zeros(shape)
  init_w = jnp.zeros(shape)

  _, _, done, u, w = lax.while_loop(
    cond_fun=cond_fn,
    body_fun=body_fn,
    init_val=(jnp.array(0), key, init_done, init_u, init_w),
  )

  return jnp.sign(u) * jnp.arccos(w)


def _loc_scale(loc, scale, value):
  if loc is None:
    if scale is None:
      return JaxArray(value)
    else:
      return JaxArray(value * scale)
  else:
    if scale is None:
      return JaxArray(value + loc)
    else:
      return JaxArray(value * scale + loc)


def _check_py_seq(seq):
  return jnp.asarray(seq) if isinstance(seq, (tuple, list)) else seq


class RandomState(Variable):
  """RandomState that track the random generator state. """
  __slots__ = ()

  def __init__(self, seed=None):
    """RandomState constructor.

    Parameters
    ----------
    seed : int, jax.DeviceArray, Optional
      The initial seed of the random number generator.
    """
    if seed is None:
      seed = np.random.randint(0, 100000, 2, dtype=np.uint32)
    if isinstance(seed, int):
      key = jr.PRNGKey(seed)
    else:
      assert len(seed) == 2
      key = seed
    super(RandomState, self).__init__(key)

  # ------------------- #
  # seed and random key #
  # ------------------- #

  def seed(self, seed):
    """Sets a new random seed.

    Parameters
    ----------
    seed : int
      The new initial seed of the random number generator.
    """
    self.value = jr.PRNGKey(seed)

  def split_key(self):
    """Create a new seed from the current seed.
    """
    if not isinstance(self.value, jnp.ndarray):
      self._value = jnp.asarray(self.value)
    keys = jr.split(self.value, num=2)
    self._value = keys[0]
    return keys[1]

  def split_keys(self, n):
    """Create multiple seeds from the current seed. This is used
    internally by `pmap` and `vmap` to ensure that random numbers
    are different in parallel threads.

    Parameters
    ----------
    n : int
      The number of seeds to generate.
    """
    keys = jr.split(self.value, n + 1)
    self._value = keys[0]
    return keys[1:]

  # ---------------- #
  # random functions #
  # ---------------- #

  def rand(self, *dn):
    return JaxArray(jr.uniform(self.split_key(), shape=dn, minval=0., maxval=1.))

  def randint(self, low, high=None, size=None, dtype=jnp.int_):
    low = _remove_jax_array(low)
    high = _remove_jax_array(high)
    if high is None:
      high = low
      low = 0
    high = _check_py_seq(high)
    low = _check_py_seq(low)
    if size is None:
      size = lax.broadcast_shapes(jnp.shape(low),
                                  jnp.shape(high))
    return JaxArray(jr.randint(self.split_key(),
                               shape=_size2shape(size),
                               minval=low, maxval=high, dtype=dtype))

  def random_integers(self, low, high=None, size=None):
    low = _remove_jax_array(low)
    high = _remove_jax_array(high)
    low = _check_py_seq(low)
    high = _check_py_seq(high)
    if high is None:
      high = low
      low = 1
    high += 1
    if size is None:
      size = lax.broadcast_shapes(jnp.shape(low), jnp.shape(high))
    return JaxArray(jr.randint(self.split_key(),
                               shape=_size2shape(size),
                               minval=low,
                               maxval=high))

  def randn(self, *dn):
    return JaxArray(jr.normal(self.split_key(), shape=dn))

  def random(self, size=None):
    return JaxArray(jr.uniform(self.split_key(), shape=_size2shape(size), minval=0., maxval=1.))

  def random_sample(self, size=None):
    return self.random(size=size)

  def ranf(self, size=None):
    return self.random(size=size)

  def sample(self, size=None):
    return self.random(size=size)

  def choice(self, a, size=None, replace=True, p=None):
    a = _remove_jax_array(a)
    p = _remove_jax_array(p)
    a = _check_py_seq(a)
    p = _check_py_seq(p)
    return JaxArray(jr.choice(self.split_key(), a=a, shape=_size2shape(size),
                              replace=replace, p=p))

  def permutation(self, x):
    x = x.value if isinstance(x, JaxArray) else x
    x = _check_py_seq(x)
    return JaxArray(jr.permutation(self.split_key(), x))

  def shuffle(self, x, axis=0):
    assert isinstance(x, JaxArray), f'Must be a JaxArray, but got {type(x)}'
    x.value = jr.permutation(self.split_key(), x.value, axis=axis)

  def beta(self, a, b, size=None):
    a = a.value if isinstance(a, JaxArray) else a
    b = b.value if isinstance(b, JaxArray) else b
    a = _check_py_seq(a)
    b = _check_py_seq(b)
    if size is None:
      size = lax.broadcast_shapes(jnp.shape(a), jnp.shape(b))
    return JaxArray(jr.beta(self.split_key(), a=a, b=b, shape=_size2shape(size)))

  def exponential(self, scale=None, size=None):
    scale = _remove_jax_array(scale)
    scale = _check_py_seq(scale)
    if size is None:
      size = jnp.shape(scale)
    r = jr.exponential(self.split_key(), shape=_size2shape(size))
    if scale is None:
      return JaxArray(r)
    else:
      return JaxArray(r / scale)

  def gamma(self, shape, scale=None, size=None):
    shape = _remove_jax_array(shape)
    scale = _remove_jax_array(scale)
    shape = _check_py_seq(shape)
    scale = _check_py_seq(scale)
    if size is None:
      size = lax.broadcast_shapes(jnp.shape(shape), jnp.shape(scale))
    r = jr.gamma(self.split_key(), a=shape, shape=_size2shape(size))
    if scale is None:
      return JaxArray(r)
    else:
      return JaxArray(r * scale)

  def gumbel(self, loc=None, scale=None, size=None):
    loc = _remove_jax_array(loc)
    scale = _remove_jax_array(scale)
    loc = _check_py_seq(loc)
    scale = _check_py_seq(scale)
    if size is None:
      size = lax.broadcast_shapes(jnp.shape(loc), jnp.shape(scale))
    return _loc_scale(loc, scale, jr.gumbel(self.split_key(), shape=_size2shape(size)))

  def laplace(self, loc=None, scale=None, size=None):
    loc = _remove_jax_array(loc)
    scale = _remove_jax_array(scale)
    loc = _check_py_seq(loc)
    scale = _check_py_seq(scale)
    if size is None:
      size = lax.broadcast_shapes(jnp.shape(loc), jnp.shape(scale))
    return _loc_scale(loc, scale, jr.laplace(self.split_key(), shape=_size2shape(size)))

  def logistic(self, loc=None, scale=None, size=None):
    loc = _remove_jax_array(loc)
    scale = _remove_jax_array(scale)
    loc = _check_py_seq(loc)
    scale = _check_py_seq(scale)
    if size is None:
      size = lax.broadcast_shapes(jnp.shape(loc), jnp.shape(scale))
    return _loc_scale(loc, scale, jr.logistic(self.split_key(), shape=_size2shape(size)))

  def normal(self, loc=None, scale=None, size=None):
    loc = _remove_jax_array(loc)
    scale = _remove_jax_array(scale)
    loc = _check_py_seq(loc)
    scale = _check_py_seq(scale)
    if size is None:
      size = lax.broadcast_shapes(jnp.shape(scale), jnp.shape(loc))
    return _loc_scale(loc, scale, jr.normal(self.split_key(), shape=_size2shape(size)))

  def pareto(self, a, size=None):
    a = _remove_jax_array(a)
    a = _check_py_seq(a)
    if size is None:
      size = jnp.shape(a)
    return JaxArray(jr.pareto(self.split_key(), b=a, shape=_size2shape(size)))

  def poisson(self, lam=1.0, size=None):
    lam = _check_py_seq(_remove_jax_array(lam))
    if size is None:
      size = jnp.shape(lam)
    return JaxArray(jr.poisson(self.split_key(), lam=lam, shape=_size2shape(size)))

  def standard_cauchy(self, size=None):
    return JaxArray(jr.cauchy(self.split_key(), shape=_size2shape(size)))

  def standard_exponential(self, size=None):
    return JaxArray(jr.exponential(self.split_key(), shape=_size2shape(size)))

  def standard_gamma(self, shape, size=None):
    shape = _remove_jax_array(shape)
    shape = _check_py_seq(shape)
    if size is None:
      size = jnp.shape(shape)
    return JaxArray(jr.gamma(self.split_key(), a=shape, shape=_size2shape(size)))

  def standard_normal(self, size=None):
    return JaxArray(jr.normal(self.split_key(), shape=_size2shape(size)))

  def standard_t(self, df, size=None):
    df = _remove_jax_array(df)
    df = _check_py_seq(df)
    if size is None:
      size = jnp.shape(size)
    return JaxArray(jr.t(self.split_key(), df=df, shape=_size2shape(size)))

  def uniform(self, low=0.0, high=1.0, size=None):
    low = _remove_jax_array(low)
    high = _remove_jax_array(high)
    low = _check_py_seq(low)
    high = _check_py_seq(high)
    if size is None:
      size = lax.broadcast_shapes(jnp.shape(low), jnp.shape(high))
    return JaxArray(jr.uniform(self.split_key(),
                               shape=_size2shape(size),
                               minval=low,
                               maxval=high))

  def truncated_normal(self, lower, upper, size, scale=None):
    lower = _remove_jax_array(lower)
    lower = _check_py_seq(lower)
    upper = _remove_jax_array(upper)
    upper = _check_py_seq(upper)
    scale = _remove_jax_array(scale)
    scale = _check_py_seq(scale)
    if size is None:
      size = lax.broadcast_shapes(jnp.shape(lower),
                                  jnp.shape(upper),
                                  jnp.shape(scale))
    rands = jr.truncated_normal(self.split_key(),
                                lower=lower,
                                upper=upper,
                                shape=_size2shape(size))
    if scale is None:
      return JaxArray(rands)
    else:
      return JaxArray(rands * scale)

  def _check_p(self, p):
    raise ValueError(f'Parameter p should be within [0, 1], but we got {p}')

  def bernoulli(self, p, size=None):
    p = _check_py_seq(_remove_jax_array(p))
    check_error_in_jit(jnp.any(jnp.logical_and(p < 0, p > 1)),  self._check_p, p)
    if size is None:
      size = jnp.shape(p)
    return JaxArray(jr.bernoulli(self.split_key(), p=p, shape=_size2shape(size)))

  def lognormal(self, mean=None, sigma=None, size=None):
    mean = _check_py_seq(_remove_jax_array(mean))
    sigma = _check_py_seq(_remove_jax_array(sigma))
    if size is None:
      size = jnp.broadcast_shapes(jnp.shape(mean),
                                  jnp.shape(sigma))
    samples = jr.normal(self.split_key(), shape=_size2shape(size))
    samples = _loc_scale(mean, sigma, samples)
    samples = jnp.exp(samples.value)
    return JaxArray(samples)

  def binomial(self, n, p, size=None):
    n = _check_py_seq(n.value if isinstance(n, JaxArray) else n)
    p = _check_py_seq(p.value if isinstance(p, JaxArray) else p)
    check_error_in_jit(jnp.any(jnp.logical_and(p < 0, p > 1)), self._check_p, p)
    if size is None:
      size = jnp.broadcast_shapes(jnp.shape(n), jnp.shape(p))
    return JaxArray(_binomial(self.split_key(), p, n, shape=_size2shape(size)))

  def chisquare(self, df, size=None):
    df = _check_py_seq(_remove_jax_array(df))
    if size is None:
      if jnp.ndim(df) == 0:
        dist = jr.normal(self.split_key(), (df,)) ** 2
        dist = dist.sum()
      else:
        raise NotImplementedError('Do not support non-scale "df" when "size" is None')
    else:
      dist = jr.normal(self.split_key(), (df,) + _size2shape(size)) ** 2
      dist = dist.sum(axis=0)
    return JaxArray(dist)

  def dirichlet(self, alpha, size=None):
    alpha = _check_py_seq(_remove_jax_array(alpha))
    return JaxArray(jr.dirichlet(self.split_key(), alpha=alpha, shape=_size2shape(size)))

  def geometric(self, p, size=None):
    p = _remove_jax_array(p)
    p = _check_py_seq(p)
    if size is None:
      size = jnp.shape(p)
    u = jr.uniform(self.split_key(), size)
    r = jnp.floor(jnp.log1p(-u) / jnp.log1p(-p))
    return JaxArray(r)

  def _check_p2(self, p):
    raise ValueError(f'We require `sum(pvals[:-1]) <= 1`. But we got {p}')

  def multinomial(self, n, pvals, size=None):
    n = _check_py_seq(_remove_jax_array(n))
    pvals = _check_py_seq(_remove_jax_array(pvals))
    check_error_in_jit(jnp.sum(pvals[:-1]) > 1., self._check_p2, pvals)
    if isinstance(n, jax.core.Tracer):
      raise ValueError("The total count parameter `n` should not be a jax abstract array.")
    size = _size2shape(size)
    n_max = int(np.max(jax.device_get(n)))
    batch_shape = lax.broadcast_shapes(jnp.shape(pvals)[:-1], jnp.shape(n))
    return JaxArray(_multinomial(self.split_key(), pvals, n, n_max, batch_shape + size))

  def multivariate_normal(self, mean, cov, size=None, method: str = 'cholesky'):
    if method not in {'svd', 'eigh', 'cholesky'}:
      raise ValueError("method must be one of {'svd', 'eigh', 'cholesky'}")
    mean = _check_py_seq(_remove_jax_array(mean))
    cov = _check_py_seq(_remove_jax_array(cov))

    if not jnp.ndim(mean) >= 1:
      raise ValueError(f"multivariate_normal requires mean.ndim >= 1, got mean.ndim == {jnp.ndim(mean)}")
    if not jnp.ndim(cov) >= 2:
      raise ValueError(f"multivariate_normal requires cov.ndim >= 2, got cov.ndim == {jnp.ndim(cov)}")
    n = mean.shape[-1]
    if jnp.shape(cov)[-2:] != (n, n):
      raise ValueError(f"multivariate_normal requires cov.shape == (..., n, n) for n={n}, "
                       f"but got cov.shape == {jnp.shape(cov)}.")
    if size is None:
      size = lax.broadcast_shapes(mean.shape[:-1], cov.shape[:-2])
    else:
      size = _size2shape(size)
      _check_shape("normal", size, mean.shape[:-1], cov.shape[:-2])

    if method == 'svd':
      (u, s, _) = jnp.linalg.svd(cov)
      factor = u * jnp.sqrt(s[..., None, :])
    elif method == 'eigh':
      (w, v) = jnp.linalg.eigh(cov)
      factor = v * jnp.sqrt(w[..., None, :])
    else:  # 'cholesky'
      factor = jnp.linalg.cholesky(cov)
    normal_samples = jr.normal(self.split_key(), size + mean.shape[-1:])
    r = mean + jnp.einsum('...ij,...j->...i', factor, normal_samples)
    return JaxArray(r)

  def rayleigh(self, scale=1.0, size=None):
    scale = _check_py_seq(_remove_jax_array(scale))
    if size is None:
      size = jnp.shape(scale)
    x = jnp.sqrt(-2. * jnp.log(jr.uniform(self.split_key(), shape=_size2shape(size), minval=0, maxval=1)))
    return JaxArray(x * scale)

  def triangular(self, size=None):
    bernoulli_samples = jr.bernoulli(self.split_key(), p=0.5, shape=_size2shape(size))
    return JaxArray(2 * bernoulli_samples - 1)

  def vonmises(self, mu, kappa, size=None):
    mu = _check_py_seq(_remove_jax_array(mu))
    kappa = _check_py_seq(_remove_jax_array(kappa))
    if size is None:
      size = lax.broadcast_shapes(jnp.shape(mu), jnp.shape(kappa))
    size = _size2shape(size)
    samples = _von_mises_centered(self.split_key(), kappa, size)
    samples = samples + mu
    samples = (samples + jnp.pi) % (2.0 * jnp.pi) - jnp.pi
    return JaxArray(samples)

  def weibull(self, a, size=None):
    a = _check_py_seq(_remove_jax_array(a))
    if size is None:
      size = jnp.shape(a)
    else:
      if jnp.size(a) > 1:
        raise ValueError(f'"a" should be a scalar when "size" is provided. But we got {a}')
    size = _size2shape(size)
    random_uniform = jr.uniform(key=self.split_key(), shape=size, minval=0, maxval=1)
    r = jnp.power(-jnp.log1p(-random_uniform), 1.0 / a)
    return JaxArray(r)

  def weibull_min(self, a, scale=None, size=None):
    """Sample from a Weibull minimum distribution.

    Parameters
    ----------
    a: float, array_like
      The concentration parameter of the distribution.
    scale: float, array_like
      The scale parameter of the distribution.
    size: optional, int, tuple of int
      The shape added to the parameters loc and scale broadcastable shape.

    Returns
    -------
    out: array_like
      The sampling results.
    """
    a = _check_py_seq(_remove_jax_array(a))
    scale = _check_py_seq(_remove_jax_array(scale))
    if size is None:
      size = jnp.broadcast_shapes(jnp.shape(a), jnp.shape(scale))
    else:
      if jnp.size(a) > 1:
        raise ValueError(f'"a" should be a scalar when "size" is provided. But we got {a}')
    size = _size2shape(size)
    random_uniform = jr.uniform(key=self.split_key(), shape=size, minval=0, maxval=1)
    r = jnp.power(-jnp.log1p(-random_uniform), 1.0 / a)
    if scale is not None:
      r /= scale
    return JaxArray(r)

  def maxwell(self, size=None):
    shape = core.canonicalize_shape(_size2shape(size)) + (3,)
    norm_rvs = jr.normal(key=self.split_key(), shape=shape)
    return JaxArray(jnp.linalg.norm(norm_rvs, axis=-1))

  def negative_binomial(self, n, p, size=None):
    n = _check_py_seq(_remove_jax_array(n))
    p = _check_py_seq(_remove_jax_array(p))
    if size is None:
      size = lax.broadcast_shapes(jnp.shape(n), jnp.shape(p))
    size = _size2shape(size)
    logits = jnp.log(p) - jnp.log1p(-p)
    rate = self.gamma(shape=n, scale=jnp.exp(-logits), size=size)
    return JaxArray(self.poisson(lam=rate))

  def wald(self, mean, scale, size=None):
    mean = _check_py_seq(_remove_jax_array(mean))
    scale = _check_py_seq(_remove_jax_array(scale))
    if size is None:
      size = lax.broadcast_shapes(jnp.shape(mean), jnp.shape(scale))
    size = _size2shape(size)
    sampled_chi2 = jnp.square(self.randn(*size).value)
    sampled_uniform = self.uniform(size=size).value
    # Wikipedia defines an intermediate x with the formula
    #   x = loc + loc ** 2 * y / (2 * conc)
    #       - loc / (2 * conc) * sqrt(4 * loc * conc * y + loc ** 2 * y ** 2)
    # where y ~ N(0, 1)**2 (sampled_chi2 above) and conc is the concentration.
    # Let us write
    #   w = loc * y / (2 * conc)
    # Then we can extract the common factor in the last two terms to obtain
    #   x = loc + loc * w * (1 - sqrt(2 / w + 1))
    # Now we see that the Wikipedia formula suffers from catastrphic
    # cancellation for large w (e.g., if conc << loc).
    #
    # Fortunately, we can fix this by multiplying both sides
    # by 1 + sqrt(2 / w + 1).  We get
    #   x * (1 + sqrt(2 / w + 1)) =
    #     = loc * (1 + sqrt(2 / w + 1)) + loc * w * (1 - (2 / w + 1))
    #     = loc * (sqrt(2 / w + 1) - 1)
    # The term sqrt(2 / w + 1) + 1 no longer presents numerical
    # difficulties for large w, and sqrt(2 / w + 1) - 1 is just
    # sqrt1pm1(2 / w), which we know how to compute accurately.
    # This just leaves the matter of small w, where 2 / w may
    # overflow.  In the limit a w -> 0, x -> loc, so we just mask
    # that case.
    sqrt1pm1_arg = 4 * scale / (mean * sampled_chi2)  # 2 / w above
    safe_sqrt1pm1_arg = jnp.where(sqrt1pm1_arg < np.inf, sqrt1pm1_arg, 1.0)
    denominator = 1.0 + jnp.sqrt(safe_sqrt1pm1_arg + 1.0)
    ratio = jnp.expm1(0.5 * jnp.log1p(safe_sqrt1pm1_arg)) / denominator
    sampled = mean * jnp.where(sqrt1pm1_arg < np.inf, ratio, 1.0)  # x above
    res = jnp.where(sampled_uniform <= mean / (mean + sampled),
                    sampled,
                    jnp.square(mean) / sampled)
    return JaxArray(res)

  def t(self, df, size=None):
    df = _check_py_seq(_remove_jax_array(df))
    if size is None:
      size = np.shape(df)
    else:
      size = _size2shape(size)
      _check_shape("t", size, np.shape(df))
    keys = self.split_keys(2)
    n = jr.normal(keys[0], size)
    two = _const(n, 2)
    half_df = lax.div(df, two)
    g = jr.gamma(keys[1], half_df, size)
    return JaxArray(n * jnp.sqrt(half_df / g))

  def orthogonal(self, n: int, size=None):
    size = _size2shape(size)
    _check_shape("orthogonal", size)
    n = core.concrete_or_error(index, n, "The error occurred in jax.random.orthogonal()")
    z = jr.normal(self.split_key(), size + (n, n))
    q, r = jnp.linalg.qr(z)
    d = jnp.diagonal(r, 0, -2, -1)
    return JaxArray(q * jnp.expand_dims(d / abs(d), -2))

  def noncentral_chisquare(self, df, nonc, size=None):
    df = _check_py_seq(_remove_jax_array(df))
    nonc = _check_py_seq(_remove_jax_array(nonc))
    if size is None:
      size = lax.broadcast_shapes(jnp.shape(df), jnp.shape(nonc))
    size = _size2shape(size)
    i = jr.poisson(self.split_key(), 0.5 * nonc, shape=size)
    n = jr.normal(self.split_key(), shape=size) + jnp.sqrt(nonc)
    cond = jnp.greater(df, 1.0)
    df2 = jnp.where(cond, df - 1.0, df + 2.0 * i)
    chi2 = 2.0 * jr.gamma(self.split_key(), 0.5 * df2, shape=size)
    return JaxArray(jnp.where(cond, chi2 + n * n, chi2))

  def zipf(self, a, size=None):
    a = _check_py_seq(_remove_jax_array(a))
    if size is None:
      size = jnp.shape(a)
    return JaxArray(call(lambda x: np.random.zipf(x, size),
                         a,
                         result_shape=jax.ShapeDtypeStruct(size, jnp.int_)))

  def power(self, a, size=None):
    a = _check_py_seq(_remove_jax_array(a))
    if size is None:
      size = jnp.shape(a)
    size = _size2shape(size)
    return JaxArray(call(lambda a: np.random.power(a=a, size=size),
                         a, result_shape=jax.ShapeDtypeStruct(size, jnp.float_)))

  def f(self, dfnum, dfden, size=None):
    dfnum = _remove_jax_array(dfnum)
    dfden = _remove_jax_array(dfden)
    dfnum = _check_py_seq(dfnum)
    dfden = _check_py_seq(dfden)
    if size is None:
      size = jnp.broadcast_shapes(jnp.shape(dfnum), jnp.shape(dfden))
    size = _size2shape(size)
    d = {'dfnum': dfnum, 'dfden': dfden}
    return JaxArray(call(lambda x: np.random.f(dfnum=x['dfnum'],
                                               dfden=x['dfden'],
                                               size=size),
                         d,
                         result_shape=jax.ShapeDtypeStruct(size, jnp.float_)))

  def hypergeometric(self, ngood, nbad, nsample, size=None):
    ngood = _check_py_seq(_remove_jax_array(ngood))
    nbad = _check_py_seq(_remove_jax_array(nbad))
    nsample = _check_py_seq(_remove_jax_array(nsample))

    if size is None:
      size = lax.broadcast_shapes(jnp.shape(ngood),
                                  jnp.shape(nbad),
                                  jnp.shape(nsample))
    size = _size2shape(size)
    d = {'ngood': ngood, 'nbad': nbad, 'nsample': nsample}
    return JaxArray(call(lambda x: np.random.hypergeometric(ngood=x['ngood'],
                                                            nbad=x['nbad'],
                                                            nsample=x['nsample'],
                                                            size=size),
                         d, result_shape=jax.ShapeDtypeStruct(size, jnp.int_)))

  def logseries(self, p, size=None):
    p = _check_py_seq(_remove_jax_array(p))
    if size is None:
      size = jnp.shape(p)
    size = _size2shape(size)
    return JaxArray(call(lambda p: np.random.logseries(p=p, size=size),
                         p, result_shape=jax.ShapeDtypeStruct(size, jnp.int_)))

  def noncentral_f(self, dfnum, dfden, nonc, size=None):
    dfnum = _check_py_seq(_remove_jax_array(dfnum))
    dfden = _check_py_seq(_remove_jax_array(dfden))
    nonc = _check_py_seq(_remove_jax_array(nonc))
    if size is None:
      size = lax.broadcast_shapes(jnp.shape(dfnum),
                                  jnp.shape(dfden),
                                  jnp.shape(nonc))
    size = _size2shape(size)
    d = {'dfnum': dfnum, 'dfden': dfden, 'nonc': nonc}
    return JaxArray(call(lambda x: np.random.noncentral_f(dfnum=x['dfnum'],
                                                          dfden=x['dfden'],
                                                          nonc=x['nonc'],
                                                          size=size),
                         d, result_shape=jax.ShapeDtypeStruct(size, jnp.float_)))

  def loggamma(self, a, size=None):
    a = _check_py_seq(_remove_jax_array(a))
    if size is None:
      size = jnp.shape(a)
    return JaxArray(jr.loggamma(self.split_key(), a, shape=size))

  def categorical(self, logits, axis:int= -1, size=None):
    logits = _check_py_seq(_remove_jax_array(logits))
    if size is None:
      size = list(jnp.shape(logits))
      size.pop(axis)
    return JaxArray(jr.categorical(self.split_key(), logits, axis=axis, shape=size))


# alias
Generator = RandomState

# register pytree
register_pytree_node(RandomState,
                     lambda t: ((t.value,), None),
                     lambda aux_data, flat_contents: RandomState(*flat_contents))

# default random genrator
DEFAULT = RandomState(np.random.randint(0, 10000, size=2, dtype=np.uint32))


@wraps(np.random.default_rng)
def default_rng(seed=None):
  return RandomState(seed)


@wraps(np.random.seed)
def seed(seed=None):
  DEFAULT.seed(np.random.randint(0, 100000) if seed is None else seed)


@wraps(np.random.rand)
def rand(*dn):
  return DEFAULT.rand(*dn)


@wraps(np.random.randint)
def randint(low, high=None, size=None, dtype=jnp.int_):
  return DEFAULT.randint(low, high=high, size=size, dtype=dtype)


@wraps(np.random.random_integers)
def random_integers(low, high=None, size=None):
  return DEFAULT.random_integers(low, high=high, size=size)


@wraps(np.random.randn)
def randn(*dn):
  return DEFAULT.randn(*dn)


@wraps(np.random.random)
def random(size=None):
  return DEFAULT.random(size)


@wraps(np.random.random_sample)
def random_sample(size=None):
  return DEFAULT.random_sample(size)


@wraps(np.random.ranf)
def ranf(size=None):
  return DEFAULT.ranf(size)


@wraps(np.random.sample)
def sample(size=None):
  return DEFAULT.sample(size)


@wraps(np.random.choice)
def choice(a, size=None, replace=True, p=None):
  a = _remove_jax_array(a)
  return DEFAULT.choice(a=a, size=size, replace=replace, p=p)


@wraps(np.random.permutation)
def permutation(x):
  return DEFAULT.permutation(x)


@wraps(np.random.shuffle)
def shuffle(x, axis=0):
  DEFAULT.shuffle(x, axis)


@wraps(np.random.beta)
def beta(a, b, size=None):
  return DEFAULT.beta(a, b, size=size)


@wraps(np.random.exponential)
def exponential(scale=None, size=None):
  return DEFAULT.exponential(scale, size)


@wraps(np.random.gamma)
def gamma(shape, scale=None, size=None):
  return DEFAULT.gamma(shape, scale, size=size)


@wraps(np.random.gumbel)
def gumbel(loc=None, scale=None, size=None):
  return DEFAULT.gumbel(loc, scale, size=size)


@wraps(np.random.laplace)
def laplace(loc=None, scale=None, size=None):
  return DEFAULT.laplace(loc, scale, size)


@wraps(np.random.logistic)
def logistic(loc=None, scale=None, size=None):
  return DEFAULT.logistic(loc, scale, size)


@wraps(np.random.normal)
def normal(loc=None, scale=None, size=None):
  return DEFAULT.normal(loc, scale, size)


@wraps(np.random.pareto)
def pareto(a, size=None):
  return DEFAULT.pareto(a, size)


@wraps(np.random.poisson)
def poisson(lam=1.0, size=None):
  return DEFAULT.poisson(lam, size)


@wraps(np.random.standard_cauchy)
def standard_cauchy(size=None):
  return DEFAULT.standard_cauchy(size)


@wraps(np.random.standard_exponential)
def standard_exponential(size=None):
  return DEFAULT.standard_exponential(size)


@wraps(np.random.standard_gamma)
def standard_gamma(shape, size=None):
  return DEFAULT.standard_gamma(shape, size)


@wraps(np.random.standard_normal)
def standard_normal(size=None):
  return DEFAULT.standard_normal(size)


@wraps(np.random.standard_t)
def standard_t(df, size=None):
  return DEFAULT.standard_t(df, size)


@wraps(np.random.uniform)
def uniform(low=0.0, high=1.0, size=None):
  return DEFAULT.uniform(low, high, size)


@wraps(jr.truncated_normal)
def truncated_normal(lower, upper, size=None, scale=None):
  """Sample truncated standard normal random values with given shape and dtype.

  Parameters
  ----------
  lower : float, ndarray
    A float or array of floats representing the lower bound for
    truncation. Must be broadcast-compatible with ``upper``.
  upper : float, ndarray
    A float or array of floats representing the  upper bound for
    truncation. Must be broadcast-compatible with ``lower``.
  size : optional, list of int, tuple of int
    A tuple of nonnegative integers specifying the result
    shape. Must be broadcast-compatible with ``lower`` and ``upper``. The
    default (None) produces a result shape by broadcasting ``lower`` and
    ``upper``.
  scale : float, ndarray
    Standard deviation (spread or "width") of the distribution. Must be
    non-negative.

  Returns
  -------
  out : JaxArray
    A random array with the specified dtype and shape given by ``shape`` if
    ``shape`` is not None, or else by broadcasting ``lower`` and ``upper``.
    Returns values in the open interval ``(lower, upper)``.
  """
  return DEFAULT.truncated_normal(lower, upper, size, scale)


@wraps(jr.bernoulli)
def bernoulli(p=0.5, size=None):
  """Sample Bernoulli random values with given shape and mean.

  Parameters
  ----------
  p: float, array_like, optional
    A float or array of floats for the mean of the random
    variables. Must be broadcast-compatible with ``shape`` and the values
    should be within [0, 1]. Default 0.5.
  size: optional, tuple of int, int
    A tuple of nonnegative integers representing the result
    shape. Must be broadcast-compatible with ``p.shape``. The default (None)
    produces a result shape equal to ``p.shape``.

  Returns
  -------
  out: array_like
    A random array with boolean dtype and shape given by ``shape`` if ``shape``
    is not None, or else ``p.shape``.
  """
  return DEFAULT.bernoulli(p, size)


@wraps(np.random.lognormal)
def lognormal(mean=None, sigma=None, size=None):
  return DEFAULT.lognormal(mean, sigma, size)


@wraps(np.random.binomial)
def binomial(n, p, size=None):
  return DEFAULT.binomial(n, p, size)


@wraps(np.random.chisquare)
def chisquare(df, size=None):
  return DEFAULT.chisquare(df, size)


@wraps(np.random.dirichlet)
def dirichlet(alpha, size=None):
  return DEFAULT.dirichlet(alpha, size)


@wraps(np.random.geometric)
def geometric(p, size=None):
  return DEFAULT.geometric(p, size)


@wraps(np.random.f)
def f(dfnum, dfden, size=None):
  return DEFAULT.f(dfnum, dfden, size)


@wraps(np.random.hypergeometric)
def hypergeometric(ngood, nbad, nsample, size=None):
  return DEFAULT.hypergeometric(ngood, nbad, nsample, size)


@wraps(np.random.logseries)
def logseries(p, size=None):
  return DEFAULT.logseries(p, size)


@wraps(np.random.multinomial)
def multinomial(n, pvals, size=None):
  return DEFAULT.multinomial(n, pvals, size)


@wraps(np.random.multivariate_normal)
def multivariate_normal(mean, cov, size=None, method: str = 'cholesky'):
  return DEFAULT.multivariate_normal(mean, cov, size, method)


@wraps(np.random.negative_binomial)
def negative_binomial(n, p, size=None):
  return DEFAULT.negative_binomial(n, p, size)


@wraps(np.random.noncentral_chisquare)
def noncentral_chisquare(df, nonc, size=None):
  return DEFAULT.noncentral_chisquare(df, nonc, size)


@wraps(np.random.noncentral_f)
def noncentral_f(dfnum, dfden, nonc, size=None):
  return DEFAULT.noncentral_f(dfnum, dfden, nonc, size)


@wraps(np.random.power)
def power(a, size=None):
  return DEFAULT.power(a, size)


@wraps(np.random.rayleigh)
def rayleigh(scale=1.0, size=None):
  return DEFAULT.rayleigh(scale, size)


@wraps(np.random.triangular)
def triangular(size=None):
  return DEFAULT.triangular(size)


@wraps(np.random.vonmises)
def vonmises(mu, kappa, size=None):
  return DEFAULT.vonmises(mu, kappa, size)


@wraps(np.random.wald)
def wald(mean, scale, size=None):
  return DEFAULT.wald(mean, scale, size)


@wraps(np.random.weibull)
def weibull(a, size=None):
  return DEFAULT.weibull(a, size)


@wraps(jr.weibull_min)
def weibull_min(a, scale=None, size=None):
  return DEFAULT.weibull_min(a, scale, size)


@wraps(np.random.zipf)
def zipf(a, size=None):
  return DEFAULT.zipf(a, size)


@wraps(jr.maxwell)
def maxwell(size=None):
  return DEFAULT.maxwell(size)


def t(df, size=None):
  """Sample Student’s t random values.

  Parameters
  ----------
  df: float, array_like
    A float or array of floats broadcast-compatible with shape representing the parameter of the distribution.
  size: optional, int, tuple of int
    A tuple of non-negative integers specifying the result shape.
    Must be broadcast-compatible with `df`. The default (None) produces a result shape equal to `df.shape`.

  Returns
  -------
  out: array_like
    The sampled value.
  """
  return DEFAULT.t(df, size)


def orthogonal(n: int, size=None):
  """Sample uniformly from the orthogonal group `O(n)`.

  Parameters
  ----------
  n: int
     An integer indicating the resulting dimension.
  size: optional, int, tuple of int
    The batch dimensions of the result.

  Returns
  -------
  out: JaxArray
    The sampled results.
  """
  return DEFAULT.orthogonal(n, size)


def loggamma(a, size=None):
  """Sample log-gamma random values.

  Parameters
  ----------
  a: float, array_like
    A float or array of floats broadcast-compatible with shape representing the parameter of the distribution.
  size: optional, int, tuple of int
    A tuple of nonnegative integers specifying the result shape.
    Must be broadcast-compatible with `a`. The default (None) produces a result shape equal to `a.shape`.

  Returns
  -------
  out: array_like
    The sampled results.
  """
  return DEFAULT.loggamma(a, size)


@wraps(jr.categorical)
def categorical(logits, axis:int= -1, size=None):
  return DEFAULT.categorical(logits, axis, size)



# -*- coding: utf-8 -*-

import math as pm
import warnings
from typing import Union, Dict, Callable, Sequence

import jax.numpy as jnp
import numpy as np

import brainpy.math as bm
from brainpy import tools
from brainpy.base.base import Base
from brainpy.base.collector import Collector
from brainpy.connect import TwoEndConnector, MatConn, IJConn
from brainpy.errors import ModelBuildError
from brainpy.initialize import Initializer, init_param, Uniform
from brainpy.integrators import Integrator, odeint
from brainpy.tools.others import to_size, size2num
from brainpy.types import Tensor, Shape

__all__ = [
  'DynamicalSystem',
  'Container',
  'Network',
  'ConstantDelay',
  'NeuGroup',
  'ConNeuGroup',
  'TwoEndConn',
  'Channel',

  'ContainerWrapper',
]

_error_msg = 'Unknown type of the update function: {} ({}). ' \
             'Currently, BrainPy only supports: \n' \
             '1. function \n' \
             '2. function name (str) \n' \
             '3. tuple/dict of functions \n' \
             '4. tuple of function names \n'


class DynamicalSystem(Base):
  """Base Dynamical System class.

  Any object has step functions will be a dynamical system.
  That is to say, in BrainPy, the essence of the dynamical system
  is the "step functions".

  Parameters
  ----------
  name : str, optional
      The name of the dynamic system.
  """

  """Global delay variables. Useful when the same target
     variable is used in multiple mappings."""
  global_delay_vars: Dict[str, bm.LengthDelay] = Collector()

  def __init__(self, name=None):
    super(DynamicalSystem, self).__init__(name=name)

    # local delay variables
    self.local_delay_vars: Dict[str, bm.LengthDelay] = Collector()

  def __repr__(self):
    return f'{self.__class__.__name__}(name={self.name})'

  @property
  def steps(self):
    warnings.warn('.steps has been deprecated since version 2.0.3.', DeprecationWarning)
    return {}

  def ints(self, method='absolute'):
    """Collect all integrators in this node and the children nodes.

    Parameters
    ----------
    method : str
      The method to access the integrators.

    Returns
    -------
    collector : Collector
      The collection contained (the path, the integrator).
    """
    nodes = self.nodes(method=method)
    gather = Collector()
    for node_path, node in nodes.items():
      for k in dir(node):
        v = getattr(node, k)
        if isinstance(v, Integrator):
          gather[f'{node_path}.{k}' if node_path else k] = v
    return gather

  def __call__(self, *args, **kwargs):
    """The shortcut to call ``update`` methods."""
    return self.update(*args, **kwargs)

  def register_delay(
      self,
      name: str,
      delay_step: Union[int, Tensor, Callable, Initializer],
      delay_target: Union[bm.JaxArray, jnp.ndarray],
      initial_delay_data: Union[Initializer, Callable, Tensor, float, int, bool] = None,
      domain: str = 'global'
  ):
    """Register delay variable.

    Parameters
    ----------
    name: str
      The delay variable name.
    delay_step: int, JaxArray, ndarray, callable, Initializer
      The number of the steps of the delay.
    delay_target: JaxArray, ndarray, Variable
      The target for delay.
    initial_delay_data: float, int, JaxArray, ndarray, callable, Initializer
      The initializer for the delay data.
    domain: str
      The domain of the delay data to store.

    Returns
    -------
    delay_step: int, JaxArray, ndarray
      The number of the delay steps.
    """
    # delay steps
    if delay_step is None:
      return delay_step
    elif isinstance(delay_step, int):
      delay_type = 'homo'
    elif isinstance(delay_step, (bm.ndarray, jnp.ndarray, np.ndarray)):
      delay_type = 'heter'
      delay_step = bm.asarray(delay_step)
    elif callable(delay_step):
      delay_step = init_param(delay_step, delay_target.shape, allow_none=False)
      delay_type = 'heter'
    else:
      raise ValueError(f'Unknown "delay_steps" type {type(delay_step)}, only support '
                       f'integer, array of integers, callable function, brainpy.init.Initializer.')
    if delay_type == 'heter':
      if delay_step.dtype not in [bm.int32, bm.int64]:
        raise ValueError('Only support delay steps of int32, int64. If your '
                         'provide delay time length, please divide the "dt" '
                         'then provide us the number of delay steps.')
      if delay_target.shape[0] != delay_step.shape[0]:
        raise ValueError(f'Shape is mismatched: {delay_target.shape[0]} != {delay_step.shape[0]}')
    max_delay_step = int(bm.max(delay_step))

    # delay domain
    if domain not in ['global', 'local']:
      raise ValueError('"domain" must be a string in ["global", "local"]. '
                       f'Bug we got {domain}.')

    # delay variable
    if domain == 'local':
      self.local_delay_vars[name] = bm.LengthDelay(delay_target, max_delay_step, initial_delay_data)
      self.register_implicit_nodes(self.local_delay_vars)
    else:
      if name not in self.global_delay_vars:
        self.global_delay_vars[name] = bm.LengthDelay(delay_target, max_delay_step, initial_delay_data)
        # save into local delay vars when first seen "var",
        # for later update current value!
        self.local_delay_vars[name] = self.global_delay_vars[name]
      else:
        if self.global_delay_vars[name].num_delay_step - 1 < max_delay_step:
          self.global_delay_vars[name].reset(delay_target, max_delay_step, initial_delay_data)
      self.register_implicit_nodes(self.global_delay_vars)
    return delay_step

  def get_delay_data(
      self,
      name: str,
      delay_step: Union[int, bm.JaxArray, jnp.DeviceArray],
      *indices: Union[int, bm.JaxArray, jnp.DeviceArray],
  ):
    """Get delay data according to the provided delay steps.

    Parameters
    ----------
    name: str
      The delay variable name.
    delay_step: int, JaxArray, ndarray
      The delay length.
    indices: optional, int, JaxArray, ndarray
      The indices of the delay.

    Returns
    -------
    delay_data: JaxArray, ndarray
      The delay data at the given time.
    """
    if name in self.global_delay_vars:
      if isinstance(delay_step, int):
        return self.global_delay_vars[name](delay_step, *indices)
      else:
        if len(indices) == 0:
          indices = (jnp.arange(delay_step.size), )
        return self.global_delay_vars[name](delay_step, *indices)
    elif name in self.local_delay_vars:
      if isinstance(delay_step, int):
        return self.local_delay_vars[name](delay_step)
      else:
        if len(indices) == 0:
          indices = (jnp.arange(delay_step.size), )
        return self.local_delay_vars[name](delay_step, *indices)
    else:
      raise ValueError(f'{name} is not defined in delay variables.')

  def update_delay(
      self,
      name: str,
      delay_data: Union[float, bm.JaxArray, jnp.ndarray]
  ):
    """Update the delay according to the delay data.

    Parameters
    ----------
    name: str
      The name of the delay.
    delay_data: float, JaxArray, ndarray
      The delay data to update at the current time.
    """
    if name in self.local_delay_vars:
      return self.local_delay_vars[name].update(delay_data)
    else:
      if name not in self.global_delay_vars:
        raise ValueError(f'{name} is not defined in delay variables.')

  def reset_delay(
      self,
      name: str,
      delay_target: Union[bm.JaxArray, jnp.DeviceArray]
  ):
    """Reset the delay variable."""
    if name in self.local_delay_vars:
      return self.local_delay_vars[name].reset(delay_target)
    else:
      if name not in self.global_delay_vars:
        raise ValueError(f'{name} is not defined in delay variables.')

  def update(self, t, dt):
    """The function to specify the updating rule.
    Assume any dynamical system depends on the time variable ``t`` and
    the time step ``dt``.
    """
    raise NotImplementedError('Must implement "update" function by subclass self.')

  def reset(self):
    """Reset function which reset the whole variables in the model.
    """
    raise NotImplementedError('Must implement "reset" function by subclass self.')


class Container(DynamicalSystem):
  """Container object which is designed to add other instances of DynamicalSystem.

  Parameters
  ----------
  steps : tuple of function, tuple of str, dict of (str, function), optional
      The step functions.
  monitors : tuple, list, Monitor, optional
      The monitor object.
  name : str, optional
      The object name.
  show_code : bool
      Whether show the formatted code.
  ds_dict : dict of (str, )
      The instance of DynamicalSystem with the format of "key=dynamic_system".
  """

  def __init__(self, *ds_tuple, name=None, **ds_dict):
    super(Container, self).__init__(name=name)

    # children dynamical systems
    self.implicit_nodes = Collector()
    for ds in ds_tuple:
      if not isinstance(ds, DynamicalSystem):
        raise ModelBuildError(f'{self.__class__.__name__} receives instances of '
                              f'DynamicalSystem, however, we got {type(ds)}.')
      if ds.name in self.implicit_nodes:
        raise ValueError(f'{ds.name} has been paired with {ds}. Please change a unique name.')
    self.register_implicit_nodes({node.name: node for node in ds_tuple})
    for key, ds in ds_dict.items():
      if not isinstance(ds, DynamicalSystem):
        raise ModelBuildError(f'{self.__class__.__name__} receives instances of '
                              f'DynamicalSystem, however, we got {type(ds)}.')
      if key in self.implicit_nodes:
        raise ValueError(f'{key} has been paired with {ds}. Please change a unique name.')
    self.register_implicit_nodes(ds_dict)

  def __repr__(self):
    cls_name = self.__class__.__name__
    # split = '\n' + (' ' * (len(cls_name) + 1))
    split = ', '
    children = [f'{key}={str(val)}' for key, val in self.implicit_nodes.items()]
    return f'{cls_name}({split.join(children)})'

  def update(self, t, dt):
    """Step function of a network.

    In this update function, the update functions in children systems are
    iteratively called.
    """
    nodes = self.nodes(level=1, include_self=False).subset(DynamicalSystem).unique()
    for node in nodes.values():
      node.update(t, dt)

  def __getitem__(self, item):
    """Wrap the slice access (self['']). """
    if item in self.implicit_nodes:
      return self.implicit_nodes[item]
    else:
      raise ValueError(f'Unknown item {item}, we only found {list(self.implicit_nodes.keys())}')

  def __getattr__(self, item):
    """Wrap the dot access ('self.'). """
    child_ds = super(Container, self).__getattribute__('implicit_nodes')
    if item in child_ds:
      return child_ds[item]
    else:
      return super(Container, self).__getattribute__(item)

  def reset(self):
    nodes = self.nodes(level=1, include_self=False).subset(DynamicalSystem).unique()
    neuron_groups = nodes.subset(NeuGroup)
    synapse_groups = nodes.subset(TwoEndConn)
    for node in neuron_groups.values():
      node.reset()
    for node in synapse_groups.values():
      node.reset()
    for node in (nodes - neuron_groups - synapse_groups).values():
      node.reset()

  @classmethod
  def has(cls, **children_cls):
    """The aggressive operation to gather master and children classes.

    Parameters
    ----------
    children_cls
      The children classes.

    Returns
    -------
    wrapper: ContainerWrapper
      A wrapper which has master and its children classes.
    """
    return ContainerWrapper(master=cls, **children_cls)


class Network(Container):
  """Base class to model network objects, an alias of Container.

  Network instantiates a network, which is aimed to load
  neurons, synapses, and other brain objects.

  Parameters
  ----------
  name : str, Optional
    The network name.
  monitors : optional, list of str, tuple of str
    The items to monitor.
  ds_tuple :
    A list/tuple container of dynamical system.
  ds_dict :
    A dict container of dynamical system.
  """

  def __init__(self, *ds_tuple, name=None, **ds_dict):
    super(Network, self).__init__(*ds_tuple, name=name, **ds_dict)


class ConstantDelay(DynamicalSystem):
  """Class used to model constant delay variables.

  This class automatically supports batch size on the last axis. For example, if
  you run batch with the size of (10, 100), where `100` are batch size, then this
  class can automatically support your batched data.
  For examples,

  >>> import brainpy as bp
  >>> bp.dyn.ConstantDelay(size=(10, 100), delay=10.)

  This class also support nonuniform delays.

  >>> bp.dyn.ConstantDelay(size=100, delay=bp.math.random.random(100) * 4 + 10)

  Parameters
  ----------
  size : int, list of int, tuple of int
    The delay data size.
  delay : int, float, function, ndarray
    The delay time. With the unit of `dt`.
  dt: float, optional
    The time precision.
  name : optional, str
    The name of the dynamic system.
  """

  def __init__(self, size, delay, dtype=None, dt=None, **kwargs):
    # dt
    self.dt = bm.get_dt() if dt is None else dt
    self.dtype = dtype

    # data size
    if isinstance(size, int): size = (size,)
    if not isinstance(size, (tuple, list)):
      raise ModelBuildError(f'"size" must a tuple/list of int, but we got {type(size)}: {size}')
    self.size = tuple(size)

    # delay time length
    self.delay = delay

    # data and operations
    if isinstance(delay, (int, float)):  # uniform delay
      self.uniform_delay = True
      self.num_step = int(pm.ceil(delay / self.dt)) + 1
      self.out_idx = bm.Variable(bm.array([0], dtype=bm.uint32))
      self.in_idx = bm.Variable(bm.array([self.num_step - 1], dtype=bm.uint32))
      self.data = bm.Variable(bm.zeros((self.num_step,) + self.size, dtype=dtype))
      self.num = 1

    else:  # non-uniform delay
      self.uniform_delay = False
      if not len(self.size) == 1:
        raise NotImplementedError(f'Currently, BrainPy only supports 1D heterogeneous '
                                  f'delays, while we got the heterogeneous delay with '
                                  f'{len(self.size)}-dimensions.')
      self.num = tools.size2num(size)
      if bm.ndim(delay) != 1:
        raise ModelBuildError(f'Only support a 1D non-uniform delay. '
                              f'But we got {delay.ndim}D: {delay}')
      if delay.shape[0] != self.size[0]:
        raise ModelBuildError(f"The first shape of the delay time size must "
                              f"be the same with the delay data size. But "
                              f"we got {delay.shape[0]} != {self.size[0]}")
      delay = bm.around(delay / self.dt)
      self.diag = bm.array(bm.arange(self.num), dtype=bm.int_)
      self.num_step = bm.array(delay, dtype=bm.uint32) + 1
      self.in_idx = bm.Variable(self.num_step - 1)
      self.out_idx = bm.Variable(bm.zeros(self.num, dtype=bm.uint32))
      self.data = bm.Variable(bm.zeros((self.num_step.max(),) + size, dtype=dtype))

    super(ConstantDelay, self).__init__(**kwargs)

  def reset(self):
    """Reset the variables."""
    self.in_idx[:] = self.num_step - 1
    self.out_idx[:] = 0
    self.data[:] = 0

  @property
  def oldest(self):
    return self.pull()

  @property
  def latest(self):
    if self.uniform_delay:
      return self.data[self.in_idx[0]]
    else:
      return self.data[self.in_idx, self.diag]

  def pull(self):
    if self.uniform_delay:
      return self.data[self.out_idx[0]]
    else:
      return self.data[self.out_idx, self.diag]

  def push(self, value):
    if self.uniform_delay:
      self.data[self.in_idx[0]] = value
    else:
      self.data[self.in_idx, self.diag] = value

  def update(self, t=None, dt=None, **kwargs):
    """Update the delay index."""
    self.in_idx[:] = (self.in_idx + 1) % self.num_step
    self.out_idx[:] = (self.out_idx + 1) % self.num_step


class NeuGroup(DynamicalSystem):
  """Base class to model neuronal groups.

  There are several essential attributes:

  - ``size``: the geometry of the neuron group. For example, `(10, )` denotes a line of
    neurons, `(10, 10)` denotes a neuron group aligned in a 2D space, `(10, 15, 4)` denotes
    a 3-dimensional neuron group.
  - ``num``: the flattened number of neurons in the group. For example, `size=(10, )` => \
    `num=10`, `size=(10, 10)` => `num=100`, `size=(10, 15, 4)` => `num=600`.

  Parameters
  ----------
  size : int, tuple of int, list of int
    The neuron group geometry.
  name : optional, str
    The name of the dynamic system.
  """

  def __init__(self,
               size: Shape,
               name: str = None):
    # size
    if isinstance(size, (list, tuple)):
      if len(size) <= 0:
        raise ModelBuildError(f'size must be int, or a tuple/list of int. '
                              f'But we got {type(size)}')
      if not isinstance(size[0], int):
        raise ModelBuildError('size must be int, or a tuple/list of int.'
                              f'But we got {type(size)}')
      size = tuple(size)
    elif isinstance(size, int):
      size = (size,)
    else:
      raise ModelBuildError('size must be int, or a tuple/list of int.'
                            f'But we got {type(size)}')
    self.size = size
    self.num = tools.size2num(size)

    # initialize
    super(NeuGroup, self).__init__(name=name)

  def update(self, t, dt):
    """The function to specify the updating rule.

    Parameters
    ----------
    t : float
      The current time.
    dt : float
      The time step.
    """
    raise NotImplementedError(f'Subclass of {self.__class__.__name__} must '
                              f'implement "update" function.')



class TwoEndConn(DynamicalSystem):
  """Base class to model two-end synaptic connections.

  Parameters
  ----------
  pre : NeuGroup
      Pre-synaptic neuron group.
  post : NeuGroup
      Post-synaptic neuron group.
  conn : optional, ndarray, JaxArray, dict, TwoEndConnector
      The connection method between pre- and post-synaptic groups.
  name : str, optional
      The name of the dynamic system.
  """

  def __init__(
      self,
      pre: NeuGroup,
      post: NeuGroup,
      conn: Union[TwoEndConnector, Tensor, Dict[str, Tensor]] = None,
      name: str = None
  ):

    # pre or post neuron group
    # ------------------------
    if not isinstance(pre, NeuGroup):
      raise ModelBuildError('"pre" must be an instance of NeuGroup.')
    if not isinstance(post, NeuGroup):
      raise ModelBuildError('"post" must be an instance of NeuGroup.')
    self.pre = pre
    self.post = post

    # connectivity
    # ------------
    if isinstance(conn, TwoEndConnector):
      self.conn = conn(pre.size, post.size)
    elif isinstance(conn, (bm.ndarray, np.ndarray, jnp.ndarray)):
      if (pre.num, post.num) != conn.shape:
        raise ModelBuildError(f'"conn" is provided as a matrix, and it is expected '
                              f'to be an array with shape of (pre.num, post.num) = '
                              f'{(pre.num, post.num)}, however we got {conn.shape}')
      self.conn = MatConn(conn_mat=conn)
    elif isinstance(conn, dict):
      if not ('i' in conn and 'j' in conn):
        raise ModelBuildError(f'"conn" is provided as a dict, and it is expected to '
                              f'be a dictionary with "i" and "j" specification, '
                              f'however we got {conn}')
      self.conn = IJConn(i=conn['i'], j=conn['j'])
    elif isinstance(conn, str):
      self.conn = conn
    elif conn is None:
      self.conn = None
    else:
      raise ModelBuildError(f'Unknown "conn" type: {conn}')

    # initialize
    # ----------
    super(TwoEndConn, self).__init__(name=name)

  def check_pre_attrs(self, *attrs):
    """Check whether pre group satisfies the requirement."""
    if not hasattr(self, 'pre'):
      raise ModelBuildError('Please call __init__ function first.')
    for attr in attrs:
      if not isinstance(attr, str):
        raise ValueError(f'Must be string. But got {attr}.')
      if not hasattr(self.pre, attr):
        raise ModelBuildError(f'{self} need "pre" neuron group has attribute "{attr}".')

  def check_post_attrs(self, *attrs):
    """Check whether post group satisfies the requirement."""
    if not hasattr(self, 'post'):
      raise ModelBuildError('Please call __init__ function first.')
    for attr in attrs:
      if not isinstance(attr, str):
        raise ValueError(f'Must be string. But got {attr}.')
      if not hasattr(self.post, attr):
        raise ModelBuildError(f'{self} need "pre" neuron group has attribute "{attr}".')


class Channel(DynamicalSystem):
  """Abstract channel model."""

  def __init__(
      self,
      size: Union[int, Sequence[int]],
      name: str = None,
  ):
    super(Channel, self).__init__(name=name)
    self.size = to_size(size)
    self.num = size2num(self.size)

  def update(self, t, dt):
    raise NotImplementedError('Must be implemented by the subclass.')

  def current(self):
    raise NotImplementedError('Must be implemented by the subclass.')

  def reset(self):
    raise NotImplementedError('Must be implemented by the subclass.')


class ConNeuGroup(NeuGroup, Container):
  """Base class to model conductance-based neuron group.

  The standard formulation for a conductance-based model is given as

  .. math::

      C_m {dV \over dt} = \sum_jg_j(E - V) + I_{ext}

  where :math:`g_j=\bar{g}_{j} M^x N^y` is the channel conductance, :math:`E` is the
  reversal potential, :math:`M` is the activation variable, and :math:`N` is the
  inactivation variable.

  :math:`M` and :math:`N` have the dynamics of

  .. math::

      {dx \over dt} = \phi_x {x_\infty (V) - x \over \tau_x(V)}

  where :math:`x \in [M, N]`, :math:`\phi_x` is a temperature-dependent factor,
  :math:`x_\infty` is the steady state, and :math:`\tau_x` is the time constant.
  Equivalently, the above equation can be written as:

  .. math::

      \frac{d x}{d t}=\phi_{x}\left(\alpha_{x}(1-x)-\beta_{x} x\right)

  where :math:`\alpha_{x}` and :math:`\beta_{x}` are rate constants.

  .. versionadded:: 2.1.9

  Parameters
  ----------
  size : int, sequence of int
    The network size of this neuron group.
  method: str
    The numerical integration method.
  name : optional, str
    The neuron group name.

  """

  def __init__(
      self,
      size: Shape,
      C: Union[float, Tensor, Initializer, Callable] = 1.,
      A: Union[float, Tensor, Initializer, Callable] = 1e-3,
      V_th: Union[float, Tensor, Initializer, Callable] = 0.,
      V_initializer: Union[Initializer, Callable, Tensor] = Uniform(-70, -60.),
      method: str = 'exp_auto',
      name: str = None,
      **channels
  ):
    NeuGroup.__init__(self, size)
    Container.__init__(self, **channels, name=name)

    # parameters for neurons
    self.C = C
    self.A = A
    self.V_th = V_th
    self._V_initializer = V_initializer

    # variables
    self.V = bm.Variable(init_param(V_initializer, self.num, allow_none=False))
    self.input = bm.Variable(bm.zeros(self.num))
    self.spike = bm.Variable(bm.zeros(self.num, dtype=bool))

    # function
    self.integral = odeint(self.derivative, method=method)

  def reset(self):
    self.V.value = init_param(self._V_initializer, self.num, allow_none=False)
    self.spike[:] = False
    self.input[:] = 0

  def derivative(self, V, t):
    Iext = self.input.value * (1e-3 / self.A)
    for ch in self.implicit_nodes.values():
      Iext += ch.current(V)
    return Iext / self.C

  def update(self, t, dt):
    V = self.integral(self.V.value, t, dt)
    for node in self.implicit_nodes.unique().values():
      node.update(t, dt, self.V.value)
    self.spike.value = bm.logical_and(V >= self.V_th, self.V < self.V_th)
    self.input[:] = 0.
    self.V.value = V


class ContainerWrapper(object):
  def __init__(self, master, **children):
    self.master = master
    self.children_cls = children

    if not isinstance(master, type):
      raise TypeError(f'"master" should be a type. But we got {master}')
    # if not issubclass(master, Channel):
    #   raise TypeError(f'{master} should be a subclass of {Channel.__name__}.')
    for key, child in children.items():
      if isinstance(child, type):
        if not issubclass(child, Channel):
          raise TypeError(f'{child} should be a subclass of Base.')
        if child.master_cls is None:
          raise TypeError(f'{child} should set its master_cls.')
        if not issubclass(master, child.master_cls):
          raise TypeError(f'Type does not match. {child} requires a master with type '
                          f'of {child.master_cls}, but the master now is {master}.')
      elif isinstance(child, ContainerWrapper):
        if not issubclass(child.master, Channel):
          raise TypeError(f'{child.master} should be a subclass of Base.')
        if child.master.master_cls is None:
          raise TypeError(f'{child.master} should set its master_cls.')
        if not issubclass(master, child.master.master_cls):
          raise TypeError(f'Type does not match. {child.master} requires a master with type '
                          f'of {child.master.master_cls}, but the master now is {master}.')

      else:
        raise TypeError(f'The item in children should be a type or '
                        f'{ContainerWrapper.__name__} instance. But we got {child}')

  def __call__(self, size, *shared_args, shared_kwargs=None, **idv_args):
    if shared_kwargs is None:
      shared_kwargs = dict()

    # initialize children classes
    children = dict()
    for key, cls in self.children_cls.items():
      if key in idv_args:
        pars = idv_args.pop(key)
      else:
        pars = dict()
      children[key] = cls(size, *shared_args, **shared_kwargs, **pars)

    # initialize master class
    master = self.master(size, *shared_args, **shared_kwargs, **idv_args, **children)

    # assign master or parent to children
    for child in children.values():
      child.master = master

    return master

  def __repr__(self):
    children = [f'{key}={val.__name__}' for key, val in self.children_cls.items()]
    return f'{self.master.__name__}({", ".join(children)})'



# -*- coding: utf-8 -*-

__version__ = "2.1.12"


try:
  import jaxlib
  del jaxlib
except ModuleNotFoundError:
  raise ModuleNotFoundError(
    'Please install jaxlib. See '
    'https://brainpy.readthedocs.io/en/latest/quickstart/installation.html#dependency-2-jax '
    'for installation instructions.'
  )


# fundamental modules
from . import errors, tools, check


# "base" module
from . import base
from .base.base import Base
from .base.collector import Collector, TensorCollector


# math foundation
from . import math


# toolboxes
from . import connect, initialize, optimizers, measure, losses, datasets, inputs


# numerical integrators
from . import integrators
from .integrators import ode
from .integrators import sde
from .integrators import fde
from .integrators.ode import odeint
from .integrators.sde import sdeint
from .integrators.fde import fdeint
from .integrators.joint_eq import JointEq


# dynamics simulation
from . import dyn


# neural networks modeling
from . import nn


# running
from . import running


# automatic dynamics analysis
from . import analysis


# "visualization" module, will be removed soon
from .visualization import visualize


# compatible interface
from .compat import *  # compat


# convenient access
conn = connect
init = initialize
optim = optimizers



# -*- coding: utf-8 -*-


from typing import Union, Sequence, Any, Dict, Callable, Optional

import jax.numpy as jnp
from jax import lax
from jax.tree_util import tree_flatten, tree_unflatten

try:
  from jax.errors import UnexpectedTracerError
except ImportError:
  from jax.core import UnexpectedTracerError

from brainpy import errors
from brainpy.math.jaxarray import (JaxArray, Variable,
                                   turn_on_global_jit,
                                   turn_off_global_jit)
from brainpy.math.numpy_ops import as_device_array

__all__ = [
  'make_loop',
  'make_while',
  'make_cond',

  'cond',
  'ifelse',
  'for_loop',
  'while_loop',
]


def _get_scan_info(f, dyn_vars, out_vars=None, has_return=False):
  # iterable variables
  if isinstance(dyn_vars, dict):
    dyn_vars = tuple(dyn_vars.values())
  elif isinstance(dyn_vars, (tuple, list)):
    dyn_vars = tuple(dyn_vars)
  else:
    raise ValueError(
      f'"dyn_vars" does not support {type(dyn_vars)}, '
      f'only support dict/list/tuple of {JaxArray.__name__}')
  for v in dyn_vars:
    if not isinstance(v, JaxArray):
      raise ValueError(
        f'brainpy.math.jax.make_loop only support '
        f'{JaxArray.__name__}, but got {type(v)}')

  # outputs
  if out_vars is None:
    out_vars = ()
    _, tree = tree_flatten(out_vars)
  elif isinstance(out_vars, JaxArray):
    _, tree = tree_flatten(out_vars)
    out_vars = (out_vars,)
  elif isinstance(out_vars, dict):
    _, tree = tree_flatten(out_vars)
    out_vars = tuple(out_vars.values())
  elif isinstance(out_vars, (tuple, list)):
    _, tree = tree_flatten(out_vars)
    out_vars = tuple(out_vars)
  else:
    raise ValueError(
      f'"out_vars" does not support {type(out_vars)}, '
      f'only support dict/list/tuple of {JaxArray.__name__}')

  # functions
  if has_return:
    def fun2scan(dyn_values, x):
      for v, d in zip(dyn_vars, dyn_values): v._value = d
      results = f(x)
      dyn_values = [v.value for v in dyn_vars]
      out_values = [v.value for v in out_vars]
      return dyn_values, (out_values, results)

  else:
    def fun2scan(dyn_values, x):
      for v, d in zip(dyn_vars, dyn_values): v._value = d
      f(x)
      dyn_values = [v.value for v in dyn_vars]
      out_values = [v.value for v in out_vars]
      return dyn_values, out_values

  return fun2scan, dyn_vars, tree


def make_loop(body_fun, dyn_vars, out_vars=None, has_return=False):
  """Make a for-loop function, which iterate over inputs.

  Examples
  --------

  >>> import brainpy.math as bm
  >>>
  >>> a = bm.Variable(bm.zeros(1))
  >>> def f(x): a.value += 1.
  >>> loop = bm.make_loop(f, dyn_vars=[a], out_vars=a)
  >>> loop(bm.arange(10))
  Variable([[ 1.],
            [ 2.],
            [ 3.],
            [ 4.],
            [ 5.],
            [ 6.],
            [ 7.],
            [ 8.],
            [ 9.],
            [10.]], dtype=float32)
  >>> b = bm.Variable(bm.zeros(1))
  >>> def f(x):
  >>>   b.value += 1
  >>>   return b + 1
  >>> loop = bm.make_loop(f, dyn_vars=[b], out_vars=b, has_return=True)
  >>> hist_b, hist_b_plus = loop(bm.arange(10))
  >>> hist_b
  Variable([[ 1.],
            [ 2.],
            [ 3.],
            [ 4.],
            [ 5.],
            [ 6.],
            [ 7.],
            [ 8.],
            [ 9.],
            [10.]], dtype=float32)
  >>> hist_b_plus
  JaxArray([[ 2.],
            [ 3.],
            [ 4.],
            [ 5.],
            [ 6.],
            [ 7.],
            [ 8.],
            [ 9.],
            [10.],
            [11.]], dtype=float32)

  Parameters
  ----------
  body_fun : callable, function
    A function receive one argument. This argument refers to the iterable input ``x``.
  dyn_vars : dict of JaxArray, sequence of JaxArray
    The dynamically changed variables, while iterate between trials.
  out_vars : optional, JaxArray, dict of JaxArray, sequence of JaxArray
    The variables to output their values.
  has_return : bool
    The function has the return values.

  Returns
  -------
  loop_func : callable, function
    The function for loop iteration. This function receives one argument ``xs``, denoting
    the input tensor which interate over the time (note ``body_fun`` receive ``x``).
  """

  fun2scan, dyn_vars, tree = _get_scan_info(f=body_fun,
                                            dyn_vars=dyn_vars,
                                            out_vars=out_vars,
                                            has_return=has_return)

  # functions
  if has_return:
    def call(xs=None, length=None):
      init_values = [v.value for v in dyn_vars]
      try:
        turn_on_global_jit()
        dyn_values, (out_values, results) = lax.scan(
          f=fun2scan, init=init_values, xs=xs, length=length)
        turn_off_global_jit()
      except UnexpectedTracerError as e:
        turn_off_global_jit()
        for v, d in zip(dyn_vars, init_values): v._value = d
        raise errors.JaxTracerError(variables=dyn_vars) from e
      for v, d in zip(dyn_vars, dyn_values): v._value = d
      return tree_unflatten(tree, out_values), results

  else:
    def call(xs):
      init_values = [v.value for v in dyn_vars]
      try:
        turn_on_global_jit()
        dyn_values, out_values = lax.scan(f=fun2scan, init=init_values, xs=xs)
        turn_off_global_jit()
      except UnexpectedTracerError as e:
        turn_off_global_jit()
        for v, d in zip(dyn_vars, init_values): v._value = d
        raise errors.JaxTracerError(variables=dyn_vars) from e
      except Exception as e:
        turn_off_global_jit()
        for v, d in zip(dyn_vars, init_values): v._value = d
        raise e
      for v, d in zip(dyn_vars, dyn_values): v._value = d
      return tree_unflatten(tree, out_values)

  return call


def make_while(cond_fun, body_fun, dyn_vars):
  """Make a while-loop function.

  This function is similar to the ``jax.lax.while_loop``. The difference is that,
  if you are using ``JaxArray`` in your while loop codes, this function will help
  you make a easy while loop function. Note: ``cond_fun`` and ``body_fun`` do no
  receive any arguments. ``cond_fun`` shoule return a boolean value. ``body_fun``
  does not support return values.

  Examples
  --------
  >>> import brainpy.math as bm
  >>>
  >>> a = bm.zeros(1)
  >>>
  >>> def cond_f(x): return a[0] < 10
  >>> def body_f(x): a.value += 1.
  >>>
  >>> loop = bm.make_while(cond_f, body_f, dyn_vars=[a])
  >>> loop()
  >>> a
  JaxArray([10.], dtype=float32)

  Parameters
  ----------
  cond_fun : function, callable
    A function receives one argument, but return a boolean value.
  body_fun : function, callable
    A function receives one argument, without any returns.
  dyn_vars : dict of JaxArray, sequence of JaxArray
    The dynamically changed variables, while iterate between trials.

  Returns
  -------
  loop_func : callable, function
      The function for loop iteration, which receive one argument ``x`` for external input.
  """
  # iterable variables
  if isinstance(dyn_vars, dict):
    dyn_vars = tuple(dyn_vars.values())
  elif isinstance(dyn_vars, (tuple, list)):
    dyn_vars = tuple(dyn_vars)
  else:
    raise ValueError(f'"dyn_vars" does not support {type(dyn_vars)}, '
                     f'only support dict/list/tuple of {JaxArray.__name__}')
  for v in dyn_vars:
    if not isinstance(v, JaxArray):
      raise ValueError(f'Only support {JaxArray.__name__}, but got {type(v)}')

  def _body_fun(op):
    dyn_values, static_values = op
    for v, d in zip(dyn_vars, dyn_values): v._value = d
    body_fun(static_values)
    return [v.value for v in dyn_vars], static_values

  def _cond_fun(op):
    dyn_values, static_values = op
    for v, d in zip(dyn_vars, dyn_values): v._value = d
    return as_device_array(cond_fun(static_values))

  def call(x=None):
    dyn_init = [v.value for v in dyn_vars]
    try:
      turn_on_global_jit()
      dyn_values, _ = lax.while_loop(cond_fun=_cond_fun,
                                     body_fun=_body_fun,
                                     init_val=(dyn_init, x))
      turn_off_global_jit()
    except UnexpectedTracerError as e:
      turn_off_global_jit()
      for v, d in zip(dyn_vars, dyn_init): v._value = d
      raise errors.JaxTracerError(variables=dyn_vars) from e
    except Exception as e:
      turn_off_global_jit()
      for v, d in zip(dyn_vars, dyn_init): v._value = d
      raise e
    for v, d in zip(dyn_vars, dyn_values): v._value = d

  return call


def make_cond(true_fun, false_fun, dyn_vars=None):
  """Make a condition (if-else) function.

  Examples
  --------

  >>> import brainpy.math as bm
  >>> a = bm.zeros(2)
  >>> b = bm.ones(2)
  >>>
  >>> def true_f(x):  a.value += 1
  >>> def false_f(x): b.value -= 1
  >>>
  >>> cond = bm.make_cond(true_f, false_f, dyn_vars=[a, b])
  >>> cond(True)
  >>> a, b
  (JaxArray([1., 1.], dtype=float32),
   JaxArray([1., 1.], dtype=float32))
  >>> cond(False)
  >>> a, b
  (JaxArray([1., 1.], dtype=float32),
   JaxArray([0., 0.], dtype=float32))

  Parameters
  ----------
  true_fun : callable, function
    A function receives one argument, without any returns.
  false_fun : callable, function
    A function receives one argument, without any returns.
  dyn_vars : dict of JaxArray, sequence of JaxArray
    The dynamically changed variables.

  Returns
  -------
  cond_func : callable, function
      The condictional function receives two arguments: ``pred`` for true/false judgement
      and ``x`` for external input.
  """
  # iterable variables
  if dyn_vars is None:
    dyn_vars = []
  if isinstance(dyn_vars, JaxArray):
    dyn_vars = (dyn_vars,)
  elif isinstance(dyn_vars, dict):
    dyn_vars = tuple(dyn_vars.values())
  elif isinstance(dyn_vars, (tuple, list)):
    dyn_vars = tuple(dyn_vars)
  else:
    raise ValueError(f'"dyn_vars" does not support {type(dyn_vars)}, '
                     f'only support dict/list/tuple of {JaxArray.__name__}')
  for v in dyn_vars:
    if not isinstance(v, JaxArray):
      raise ValueError(f'Only support {JaxArray.__name__}, but got {type(v)}')

  if len(dyn_vars) > 0:
    def _true_fun(op):
      dyn_vals, static_vals = op
      for v, d in zip(dyn_vars, dyn_vals): v._value = d
      res = true_fun(static_vals)
      dyn_vals = [v.value for v in dyn_vars]
      return dyn_vals, res

    def _false_fun(op):
      dyn_vals, static_vals = op
      for v, d in zip(dyn_vars, dyn_vals): v._value = d
      res = false_fun(static_vals)
      dyn_vals = [v.value for v in dyn_vars]
      return dyn_vals, res

    def call(pred, x=None):
      old_values = [v.value for v in dyn_vars]
      try:
        turn_on_global_jit()
        dyn_values, res = lax.cond(pred, _true_fun, _false_fun, (old_values, x))
        turn_off_global_jit()
      except UnexpectedTracerError as e:
        turn_off_global_jit()
        for v, d in zip(dyn_vars, old_values): v._value = d
        raise errors.JaxTracerError(variables=dyn_vars) from e
      except Exception as e:
        turn_off_global_jit()
        for v, d in zip(dyn_vars, old_values): v._value = d
        raise e
      for v, d in zip(dyn_vars, dyn_values): v._value = d
      return res

  else:
    def call(pred, x=None):
      turn_on_global_jit()
      res = lax.cond(pred, true_fun, false_fun, x)
      turn_off_global_jit()
      return res

  return call


def _check_f(f):
  if callable(f):
    return f
  else:
    return (lambda _: f)


def _check_sequence(a):
  return isinstance(a, (list, tuple))


def cond(
    pred: bool,
    true_fun: Union[Callable, jnp.ndarray, JaxArray, float, int, bool],
    false_fun: Union[Callable, jnp.ndarray, JaxArray, float, int, bool],
    operands: Any,
    dyn_vars: Union[Variable, Sequence[Variable], Dict[str, Variable]] = None
):
  """Simple conditional statement (if-else) with instance of :py:class:`~.Variable`.

  >>> import brainpy.math as bm
  >>> a = bm.Variable(bm.zeros(2))
  >>> b = bm.Variable(bm.ones(2))
  >>> def true_f(_):  a.value += 1
  >>> def false_f(_): b.value -= 1
  >>>
  >>> bm.cond(True, true_f, false_f, dyn_vars=[a, b])
  >>> a, b
  Variable([1., 1.], dtype=float32), Variable([1., 1.], dtype=float32)
  >>>
  >>> bm.cond(False, true_f, false_f, dyn_vars=[a, b])
  >>> a, b
  Variable([1., 1.], dtype=float32), Variable([0., 0.], dtype=float32)

  Parameters
  ----------
  pred: bool
    Boolean scalar type, indicating which branch function to apply.
  true_fun: callable, jnp.ndarray, JaxArray, float, int, bool
    Function to be applied if ``pred`` is True.
    This function must receive one arguement for ``operands``.
  false_fun: callable, jnp.ndarray, JaxArray, float, int, bool
    Function to be applied if ``pred`` is False.
    This function must receive one arguement for ``operands``.
  operands: Any
    Operands (A) input to branching function depending on ``pred``. The type
    can be a scalar, array, or any pytree (nested Python tuple/list/dict) thereof.
  dyn_vars: optional, Variable, sequence of Variable, dict
    The dynamically changed variables.

  Returns
  -------
  res: Any
    The conditional results.
  """
  # checking
  true_fun = _check_f(true_fun)
  false_fun = _check_f(false_fun)
  if dyn_vars is None:
    dyn_vars = tuple()
  elif isinstance(dyn_vars, Variable):
    dyn_vars = (dyn_vars,)
  elif isinstance(dyn_vars, dict):
    dyn_vars = tuple(dyn_vars.values())
  elif isinstance(dyn_vars, (tuple, list)):
    dyn_vars = tuple(dyn_vars)
  else:
    raise ValueError(f'"dyn_vars" does not support {type(dyn_vars)}, '
                     f'only support dict/list/tuple of {Variable.__name__}')
  for v in dyn_vars:
    if not isinstance(v, Variable):
      raise ValueError(f'Only support {Variable.__name__}, but got {type(v)}')

  # calling the model
  if len(dyn_vars) > 0:
    def _true_fun(op):
      dyn_vals, static_vals = op
      for v, d in zip(dyn_vars, dyn_vals): v._value = d
      res = true_fun(static_vals)
      dyn_vals = [v.value for v in dyn_vars]
      return dyn_vals, res

    def _false_fun(op):
      dyn_vals, static_vals = op
      for v, d in zip(dyn_vars, dyn_vals): v._value = d
      res = false_fun(static_vals)
      dyn_vals = [v.value for v in dyn_vars]
      return dyn_vals, res

    old_values = [v.value for v in dyn_vars]
    try:
      turn_on_global_jit()
      dyn_values, res = lax.cond(pred=pred,
                                 true_fun=_true_fun,
                                 false_fun=_false_fun,
                                 operand=(old_values, operands))
      turn_off_global_jit()
    except UnexpectedTracerError as e:
      turn_off_global_jit()
      for v, d in zip(dyn_vars, old_values): v._value = d
      raise errors.JaxTracerError(variables=dyn_vars) from e
    except Exception as e:
      turn_off_global_jit()
      for v, d in zip(dyn_vars, old_values): v._value = d
      raise e
    for v, d in zip(dyn_vars, dyn_values): v._value = d
  else:
    turn_on_global_jit()
    res = lax.cond(pred, true_fun, false_fun, operands)
    turn_off_global_jit()
  return res


def ifelse(
    conditions: Union[bool, Sequence[bool]],
    branches: Sequence,
    operands: Any = None,
    dyn_vars: Union[Variable, Sequence[Variable], Dict[str, Variable]] = None,
    show_code: bool = False,
):
  """``If-else`` control flows looks like native Pythonic programming.

  Examples
  --------

  >>> import brainpy.math as bm
  >>> def f(a):
  >>>    return bm.ifelse(conditions=[a > 10, a > 5, a > 2, a > 0],
  >>>                     branches=[lambda _: 1,
  >>>                               lambda _: 2,
  >>>                               lambda _: 3,
  >>>                               lambda _: 4,
  >>>                               lambda _: 5])
  >>> f(1)
  4
  >>> # or, it can be expressed as:
  >>> def f(a):
  >>>   return bm.ifelse(conditions=[a > 10, a > 5, a > 2, a > 0],
  >>>                    branches=[1, 2, 3, 4, 5])
  >>> f(3)
  3

  Parameters
  ----------
  conditions: bool, sequence of bool
    The boolean conditions.
  branches: Sequence
    The branches, at least has two elements. Elements can be functions,
    arrays, or numbers. The number of ``branches`` and ``conditions`` has
    the relationship of `len(branches) == len(conditions) + 1`.
    Each branch should receive one arguement for ``operands``.
  operands: optional, Any
    The operands for each branch.
  dyn_vars: Variable, sequence of Variable, dict
    The dynamically changed variables.
  show_code: bool
    Whether show the formatted code.

  Returns
  -------
  res: Any
    The results of the control flow.
  """
  # checking
  if not isinstance(conditions, (tuple, list)):
    conditions = [conditions]
  if not isinstance(conditions, (tuple, list)):
    raise ValueError(f'"conditions" must be a tuple/list of boolean values. '
                     f'But we got {type(conditions)}: {conditions}')
  if not isinstance(branches, (tuple, list)):
    raise ValueError(f'"branches" must be a tuple/list. '
                     f'But we got {type(branches)}.')
  branches = [_check_f(b) for b in branches]
  if len(branches) != len(conditions) + 1:
    raise ValueError(f'The numbers of branches and conditions do not match. '
                     f'Got len(conditions)={len(conditions)} and len(branches)={len(branches)}. '
                     f'We expect len(conditions) + 1 == len(branches). ')
  if dyn_vars is None:
    dyn_vars = []
  if isinstance(dyn_vars, Variable):
    dyn_vars = (dyn_vars,)
  elif isinstance(dyn_vars, dict):
    dyn_vars = tuple(dyn_vars.values())
  elif isinstance(dyn_vars, (tuple, list)):
    dyn_vars = tuple(dyn_vars)
  else:
    raise ValueError(f'"dyn_vars" does not support {type(dyn_vars)}, only '
                     f'support dict/list/tuple of brainpy.math.Variable')
  for v in dyn_vars:
    if not isinstance(v, Variable):
      raise ValueError(f'Only support brainpy.math.Variable, but we got {type(v)}')

  # format new codes
  if len(conditions) == 1:
    if len(dyn_vars) > 0:
      return cond(conditions[0], branches[0], branches[1], operands, dyn_vars)
    else:
      return lax.cond(conditions[0], branches[0], branches[1], operands)
  else:
    code_scope = {'conditions': conditions, 'branches': branches}
    codes = ['def f(operands):',
             f'  f0 = branches[{len(conditions)}]']
    num_cond = len(conditions) - 1
    if len(dyn_vars) > 0:
      code_scope['_cond'] = cond
      code_scope['dyn_vars'] = dyn_vars
      for i in range(len(conditions) - 1):
        codes.append(f'  f{i + 1} = lambda r: _cond(conditions[{num_cond - i}], '
                     f'branches[{num_cond - i}], f{i}, r, dyn_vars)')
      codes.append(f'  return _cond(conditions[0], branches[0], f{len(conditions) - 1}, operands, dyn_vars)')
    else:
      code_scope['_cond'] = lax.cond
      for i in range(len(conditions) - 1):
        codes.append(f'  f{i + 1} = lambda r: _cond(conditions[{num_cond - i}], branches[{num_cond - i}], f{i}, r)')
      codes.append(f'  return _cond(conditions[0], branches[0], f{len(conditions) - 1}, operands)')
    codes = '\n'.join(codes)
    if show_code: print(codes)
    exec(compile(codes.strip(), '', 'exec'), code_scope)
    f = code_scope['f']
    return f(operands)


def for_loop(body_fun: Callable,
             dyn_vars: Union[Variable, Sequence[Variable], Dict[str, Variable]],
             operands: Any,
             reverse: bool = False,
             unroll: int = 1):
  """``for-loop`` control flow with :py:class:`~.Variable`.

  .. versionadded:: 2.1.11

  Parameters
  ----------
  body_fun: callable
    A Python function to be scanned. This function accepts one argument and returns one output.
    The argument denotes a slice of ``operands`` along its leading axis, and that
    output represents a slice of the return value.
  dyn_vars: Variable, sequence of Variable, dict
    The instances of :py:class:`~.Variable`.
  operands: Any
    The value over which to scan along the leading axis,
    where ``operands`` can be an array or any pytree (nested Python
    tuple/list/dict) thereof with consistent leading axis sizes.
  reverse: bool
    Optional boolean specifying whether to run the scan iteration
    forward (the default) or in reverse, equivalent to reversing the leading
    axes of the arrays in both ``xs`` and in ``ys``.
  unroll: int
    Optional positive int specifying, in the underlying operation of the
    scan primitive, how many scan iterations to unroll within a single
    iteration of a loop.

  Returns
  -------
  outs: Any
    The stacked outputs of ``body_fun`` when scanned over the leading axis of the inputs.
  """
  # check variables
  if isinstance(dyn_vars, Variable):
    dyn_vars = (dyn_vars,)
  elif isinstance(dyn_vars, dict):
    dyn_vars = tuple(dyn_vars.values())
  elif isinstance(dyn_vars, (tuple, list)):
    dyn_vars = tuple(dyn_vars)
  else:
    raise ValueError(f'"dyn_vars" does not support {type(dyn_vars)}, '
                     f'only support dict/list/tuple of {Variable.__name__}')
  for v in dyn_vars:
    if not isinstance(v, Variable):
      raise ValueError(f'brainpy.math.for_loop only support {Variable.__name__} in "dyn_vars", but got {type(v)}')

  # functions
  def fun2scan(dyn_vals, x):
    for v, d in zip(dyn_vars, dyn_vals): v._value = d
    results = body_fun(x)
    return [v.value for v in dyn_vars], results

  # functions
  init_vals = [v.value for v in dyn_vars]
  try:
    turn_on_global_jit()
    dyn_vals, out_vals = lax.scan(f=fun2scan, init=init_vals, xs=operands, reverse=reverse, unroll=unroll)
    turn_off_global_jit()
  except UnexpectedTracerError as e:
    turn_off_global_jit()
    for v, d in zip(dyn_vars, init_vals): v._value = d
    raise errors.JaxTracerError(variables=dyn_vars) from e
  except Exception as e:
    turn_off_global_jit()
    for v, d in zip(dyn_vars, init_vals): v._value = d
    raise e
  for v, d in zip(dyn_vars, dyn_vals): v._value = d
  return out_vals


def while_loop(
    body_fun: Callable,
    cond_fun: Callable,
    dyn_vars: Union[Variable, Sequence[Variable], Dict[str, Variable]],
    operands: Any,
):
  """``while-loop`` control flow with :py:class:`~.Variable`.

  .. versionadded:: 2.1.11

  Parameters
  ----------
  body_fun: callable
    A function which define the updating logic. It receives one argument for ``operands``, without returns.
  cond_fun: callable
    A function which define the stop condition. It receives one argument for ``operands``,
    with one boolean value return.
  dyn_vars: Variable, sequence of Variable, dict
    The dynamically changed variables.
  operands: Any
    The operands for ``body_fun`` and ``cond_fun`` functions.
  """
  # iterable variables
  if isinstance(dyn_vars, Variable):
    dyn_vars = (dyn_vars, )
  elif isinstance(dyn_vars, dict):
    dyn_vars = tuple(dyn_vars.values())
  elif isinstance(dyn_vars, (tuple, list)):
    dyn_vars = tuple(dyn_vars)
  else:
    raise ValueError(f'"dyn_vars" does not support {type(dyn_vars)}, '
                     f'only support dict/list/tuple of {Variable.__name__}')
  for v in dyn_vars:
    if not isinstance(v, Variable):
      raise ValueError(f'Only support {Variable.__name__}, but got {type(v)}')

  def _body_fun(op):
    dyn_vals, static_vals = op
    for v, d in zip(dyn_vars, dyn_vals): v._value = d
    body_fun(static_vals)
    return [v.value for v in dyn_vars], static_vals

  def _cond_fun(op):
    dyn_vals, static_vals = op
    for v, d in zip(dyn_vars, dyn_vals): v._value = d
    return as_device_array(cond_fun(static_vals))

  dyn_init = [v.value for v in dyn_vars]
  try:
    turn_on_global_jit()
    dyn_values, _ = lax.while_loop(cond_fun=_cond_fun,
                                   body_fun=_body_fun,
                                   init_val=(dyn_init, operands))
    turn_off_global_jit()
  except UnexpectedTracerError as e:
    turn_off_global_jit()
    for v, d in zip(dyn_vars, dyn_init): v._value = d
    raise errors.JaxTracerError(variables=dyn_vars) from e
  except Exception as e:
    turn_off_global_jit()
    for v, d in zip(dyn_vars, dyn_init): v._value = d
    raise e
  for v, d in zip(dyn_vars, dyn_values): v._value = d



# -*- coding: utf-8 -*-

from typing import Callable


__all__ = [
  'wraps'
]


def wraps(fun: Callable):
  """Specialized version of functools.wraps for wrapping numpy functions.

  This produces a wrapped function with a modified docstring. In particular, if
  `update_doc` is True, parameters listed in the wrapped function that are not
  supported by the decorated function will be removed from the docstring. For
  this reason, it is important that parameter names match those in the original
  numpy function.
  """
  def wrap(op):
    docstr = getattr(fun, "__doc__", None)
    op.__doc__ = docstr
    op.__wrapped__ = fun
    for attr in ['__name__', '__qualname__']:
      try:
        value = getattr(fun, attr)
      except AttributeError:
        pass
      else:
        setattr(op, attr, value)
    return op
  return wrap



# -*- coding: utf-8 -*-

from brainpy.dyn.base import Channel, ConNeuGroup

__all__ = [
  'Ion', 'IonChannel',
]


class Ion(Channel):
  """Base class for ions."""

  '''The type of the master object.'''
  master_cls = ConNeuGroup

  def update(self, t, dt, V):
    raise NotImplementedError('Must be implemented by the subclass.')

  def reset(self, V):
    raise NotImplementedError('Must be implemented by the subclass.')

  def current(self, V):
    raise NotImplementedError('Must be implemented by the subclass.')

  def __repr__(self):
    return f'{self.__class__.__name__}(size={self.size})'


class IonChannel(Channel):
  """Base class for ion channels."""

  '''The type of the master object.'''
  master_cls = ConNeuGroup

  def update(self, t, dt, V):
    raise NotImplementedError('Must be implemented by the subclass.')

  def current(self, V):
    raise NotImplementedError('Must be implemented by the subclass.')

  def reset(self, V):
    raise NotImplementedError('Must be implemented by the subclass.')

  def __repr__(self):
    return f'{self.__class__.__name__}(size={self.size})'



# -*- coding: utf-8 -*-

from brainpy.types import Shape

from brainpy.dyn.base import ConNeuGroup
from .base import IonChannel

__all__ = [
  'LeakyChannel',
  'IL',
  'IKL',
]


class LeakyChannel(IonChannel):
  """Base class for leaky channel."""
  master_cls = ConNeuGroup


class IL(LeakyChannel):
  """The leakage channel current.

  Parameters
  ----------
  g_max : float
    The leakage conductance.
  E : float
    The reversal potential.
  """

  def __init__(
      self,
      size,
      g_max=0.1,
      E=-70.,
      method: str = None,
      name: str = None,
  ):
    super(IL, self).__init__(size, name=name)

    self.E = E
    self.g_max = g_max
    self.method = method

  def reset(self, V):
    pass

  def update(self, t, dt, V):
    pass

  def current(self, V):
    return self.g_max * (self.E - V)


class IKL(IL):
  """The potassium leak channel current.

  Parameters
  ----------
  g_max : float
    The potassium leakage conductance which is modulated by both
    acetylcholine and norepinephrine.
  E : float
    The reversal potential.
  """

  def __init__(
      self,
      size: Shape,
      g_max=0.005,
      E=-90.,
      method=None,
      name=None,
  ):
    super(IKL, self).__init__(size=size, g_max=g_max, E=E, method=method, name=name)
